# flake8: noqa
from __future__ import annotations

import uuid
import warnings
from collections.abc import Collection
from dataclasses import asdict
from typing import Any, Callable, Dict, List, Literal, MutableMapping, Optional, Tuple

import numpy as np
import pandas as pd
from IPython.display import display

from outhad_edge.backend import counter
from outhad_edge.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
    tracker,
)
from outhad_edge.constants import DATETIME_UNITS
from outhad_edge.eventstream.schema import EventstreamSchema, RawDataSchema
from outhad_edge.eventstream.types import (
    EventstreamType,
    RawDataCustomColSchema,
    RawDataSchemaType,
)
from outhad_edge.preprocessing_graph import PreprocessingGraph
from outhad_edge.tooling import (
    Clusters,
    Cohorts,
    EventTimestampHist,
    Funnel,
    Sequences,
    StatTests,
    StepMatrix,
    StepSankey,
    TimedeltaHist,
    TransitionGraph,
    UserLifetimeHist,
)
from outhad_edge.tooling._describe import _Describe
from outhad_edge.tooling._describe_events import _DescribeEvents
from outhad_edge.tooling._transition_matrix import _TransitionMatrix
from outhad_edge.tooling.constants import BINS_ESTIMATORS
from outhad_edge.tooling.stattests.constants import STATTEST_NAMES
from outhad_edge.tooling.timedelta_hist.constants import (
    AGGREGATION_NAMES,
    EVENTSTREAM_GLOBAL_EVENTS,
)
from outhad_edge.tooling.typing.transition_graph import NormType, Threshold
from outhad_edge.utils import get_merged_col
from outhad_edge.utils.hash_object import hash_dataframe
from outhad_edge.utils.list import find_index

from .helpers import (
    AddNegativeEventsHelperMixin,
    AddPositiveEventsHelperMixin,
    AddStartEndEventsHelperMixin,
    CollapseLoopsHelperMixin,
    DropPathsHelperMixin,
    FilterEventsHelperMixin,
    GroupEventsBulkHelperMixin,
    GroupEventsHelperMixin,
    LabelCroppedPathsHelperMixin,
    LabelLostUsersHelperMixin,
    LabelNewUsersHelperMixin,
    PipeHelperMixin,
    RenameHelperMixin,
    SplitSessionsHelperMixin,
    TruncatePathsHelperMixin,
)
from .utils import eventstream_index

IndexOrder = List[Optional[str]]
EventsOrder = List[str]
FeatureType = Literal["tfidf", "count", "frequency", "binary", "time", "time_fraction", "external"]
NgramRange = Tuple[int, int]
Method = Literal["kmeans", "gmm"]
MetricsType = Literal["paths", "paths_share", "count", "count_share"]

DEFAULT_INDEX_ORDER: IndexOrder = [
    "profile",
    "path_start",
    "new_user",
    "existing_user",
    "cropped_left",
    "session_start",
    "session_start_cropped",
    "group_alias",
    "raw",
    "raw_sleep",
    None,
    "synthetic",
    "synthetic_sleep",
    "positive_target",
    "negative_target",
    "session_end_cropped",
    "session_end",
    "session_sleep",
    "cropped_right",
    "absent_user",
    "lost_user",
    "path_end",
]


# @TODO: проработать резервирование колонок


class Eventstream(
    CollapseLoopsHelperMixin,
    DropPathsHelperMixin,
    FilterEventsHelperMixin,
    GroupEventsHelperMixin,
    GroupEventsBulkHelperMixin,
    LabelLostUsersHelperMixin,
    AddNegativeEventsHelperMixin,
    LabelNewUsersHelperMixin,
    AddPositiveEventsHelperMixin,
    SplitSessionsHelperMixin,
    AddStartEndEventsHelperMixin,
    LabelCroppedPathsHelperMixin,
    TruncatePathsHelperMixin,
    RenameHelperMixin,
    PipeHelperMixin,
    EventstreamType,
):
    """
    Collection of tools for storing and processing clickstream data.

    Parameters
    ----------
    raw_data : pd.DataFrame or pd.Series
        Raw clickstream data.
    raw_data_schema : dict or RawDataSchema, optional
        Represents mapping rules connecting important eventstream columns with the raw data columns.
        The keys are defined in :py:class:`.RawDataSchema`. The values are the corresponding column names
        in the raw data. ``custom_cols`` key stands for the defining additional columns that can be used in
        the eventstream. See the :ref:`Eventstream user guide <eventstream_raw_data_schema>` for the details.

    schema : dict or EventstreamSchema, optional
        Represents a schema of the created eventstream. The keys are defined in
        :py:class:`.EventstreamSchema`. The values are the names of the corresponding eventstream columns.
        See the :ref:`Eventstream user guide <eventstream_field_names>` for the details.

    custom_cols : list of str, optional
        The list of additional columns from the raw data to be included in the eventstream.
        If not defined, all the columns from the raw data are included.

    prepare : bool, default True
        - If ``True``, input data will be transformed in the following way:

            * ``event_timestamp`` column is converted to pandas datetime format.
            * | ``event_type`` column is added and filled with ``raw`` value.
              | If the column exists, it remains unchanged.

        - If ``False`` - ``raw_data`` will be remained as is.

    index_order : list of str, default DEFAULT_INDEX_ORDER
        Sorting order for ``event_type`` column.
    user_sample_size : int of float, optional
        Number (``int``) or share (``float``) of all users' trajectories that will be randomly chosen
        and left in final sample (all other trajectories will be removed) .
        See :numpy_random_choice:`numpy documentation<>`.
    user_sample_seed : int, optional
        A seed value that is used to generate user samples.
        See :numpy_random_seed:`numpy documentation<>`.
    events_order : list of str, optional
        Sorting order for ``event_name`` column, if there are events with equal timestamps inside each user trajectory.
        The order of raw events is fixed once while eventstream initialization.
    add_start_end_events : bool, default True
        If True, ``path_start`` and ``path_end`` synthetic events are added to each path explicitly.
        See also :py:class:`.AddStartEndEvents` documentation.

    Notes
    -----
    See :doc:`Eventstream user guide</user_guides/eventstream>` for the details.


    """

    schema: EventstreamSchema
    events_order: EventsOrder
    index_order: IndexOrder
    __hash: str = ""
    _preprocessing_graph: PreprocessingGraph | None = None
    __clusters: Clusters | None = None

    __raw_data_schema: RawDataSchemaType
    __events: pd.DataFrame | pd.Series[Any]
    __funnel: Funnel
    __cohorts: Cohorts
    __step_matrix: StepMatrix
    __sankey: StepSankey
    __stattests: StatTests
    __sequences: Sequences
    __transition_graph: TransitionGraph
    __timedelta_hist: TimedeltaHist
    __user_lifetime_hist: UserLifetimeHist
    __event_timestamp_hist: EventTimestampHist
    __eventstream_index: int

    @time_performance(
        scope="eventstream",
        event_name="init",
    )
    def __init__(
        self,
        raw_data: pd.DataFrame | pd.Series[Any],
        raw_data_schema: RawDataSchema
        | RawDataSchemaType
        | dict[str, str | list[RawDataCustomColSchema]]
        | None = None,
        schema: EventstreamSchema | dict[str, str | list[str]] | None = None,
        prepare: bool = True,
        index_order: Optional[IndexOrder] = None,
        user_sample_size: Optional[int | float] = None,
        user_sample_seed: Optional[int] = None,
        events_order: Optional[EventsOrder] = None,
        custom_cols: List[str] | None = None,
        add_start_end_events: bool = True,
    ) -> None:
        tracking_params = dict(
            raw_data=raw_data,
            prepare=prepare,
            index_order=index_order,
            user_sample_size=user_sample_size,
            user_sample_seed=user_sample_seed,
            events_order=events_order,
            custom_cols=custom_cols,
            add_start_end_events=add_start_end_events,
        )
        not_hash_values = ["raw_data_schema", "schema"]

        if not schema:
            schema = EventstreamSchema()
        elif isinstance(schema, dict):
            schema = EventstreamSchema(**schema)  # type: ignore

        self.schema = schema
        self.__eventstream_index: int = counter.get_eventstream_index()

        if not raw_data_schema:
            raw_data_schema = RawDataSchema()
            if self.schema.event_type in raw_data.columns:
                raw_data_schema.event_type = self.schema.event_type

        elif isinstance(raw_data_schema, dict):
            raw_data_schema = RawDataSchema(**raw_data_schema)  # type: ignore
        self.__raw_data_schema = raw_data_schema

        if custom_cols is None and not self.__raw_data_schema.custom_cols and not self.schema.custom_cols:
            custom_cols = self.__define_default_custom_cols(raw_data=raw_data)

        if custom_cols and prepare:
            self.__raw_data_schema.custom_cols = []
            self.schema.custom_cols = []
            for col_name in custom_cols:
                col: RawDataCustomColSchema = {"raw_data_col": col_name, "custom_col": col_name}
                self.__raw_data_schema.custom_cols.append(col)
                self.schema.custom_cols.append(col_name)

        raw_data_schema_default_values = asdict(RawDataSchema())
        schema_default_values = asdict(EventstreamSchema())
        if isinstance(raw_data_schema, RawDataSchema):
            tracking_params["raw_data_schema"] = [
                key for key, value in raw_data_schema_default_values.items() if asdict(raw_data_schema)[key] != value
            ]
        tracking_params["schema"] = [
            key for key, value in schema_default_values.items() if asdict(self.schema)[key] != value
        ]

        self.__track_dataset(
            name="metadata",
            data=raw_data,
            params=tracking_params,
            schema=self.__raw_data_schema,
            not_hash_values=not_hash_values,
        )

        if user_sample_size is not None:
            raw_data = self.__sample_user_paths(raw_data, raw_data_schema, user_sample_size, user_sample_seed)
        if not index_order:
            self.index_order = DEFAULT_INDEX_ORDER
        else:
            self.index_order = index_order

        if events_order is not None:
            self.events_order = events_order
        else:
            self.events_order = []

        self.__events = self.__prepare_events(raw_data) if prepare else raw_data
        self.__events = self.__required_cleanup(events=self.__events)
        self.__apply_default_dataprocessors(add_start_end_events=add_start_end_events)
        self.index_events()
        if prepare:
            self.__track_dataset(
                name="metadata",
                data=self.__events,
                params=tracking_params,
                schema=self.schema,
                not_hash_values=not_hash_values,
            )

        self._preprocessing_graph = None

    @property
    def _eventstream_index(self) -> int:
        return self.__eventstream_index

    @property
    def _hash(self) -> str:
        if self.__hash == "":
            self.__hash = hash_dataframe(self.__events)
        return self.__hash

    def __track_dataset(
        self,
        name: str,
        data: pd.DataFrame | pd.Series[Any],
        params: dict[str, Any],
        schema: RawDataSchema | RawDataSchemaType | EventstreamSchema,
        not_hash_values: list[str],
    ) -> None:
        try:
            unique_users = data[schema.user_id].nunique()
        except Exception as e:
            unique_users = None

        try:
            unique_events = data[schema.event_name].nunique()
        except Exception as e:
            unique_events = None
        try:
            hist_data = data[schema.user_id].drop_duplicates()
            if len(hist_data) >= 500:
                hist_data = hist_data.sample(500, random_state=42)
            eventstream_hist = (
                data[data[schema.user_id].isin(hist_data)].groupby(schema.user_id).size().value_counts().to_dict()
            )

        except Exception:
            eventstream_hist = {}
        eventstream_hash = hash_dataframe(data=data)
        self.__hash = eventstream_hash

        performance_data: dict[str, Any] = {
            "shape": data.shape,
            "custom_cols": len(self.schema.custom_cols),
            "unique_users": unique_users,
            "unique_events": unique_events,
            "hash": self._hash,
            "eventstream_hist": eventstream_hist,
            "index": self.__eventstream_index,
        }
        collect_data_performance(
            scope="eventstream",
            event_name=name,
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data=performance_data,
            eventstream_index=self._eventstream_index,
        )

    @time_performance(
        scope="eventstream",
        event_name="copy",
    )
    def copy(self) -> Eventstream:
        """
        Make a copy of current ``eventstream``.

        Returns
        -------
        Eventstream

        """
        copied_eventstream = Eventstream(
            raw_data_schema=self.__raw_data_schema.copy(),
            raw_data=self.__events.copy(),
            schema=self.schema.copy(),
            prepare=False,
            index_order=self.index_order.copy(),
            events_order=self.events_order.copy(),
            add_start_end_events=False,
        )
        collect_data_performance(
            scope="eventstream",
            event_name="metadata",
            called_params={},
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=self._eventstream_index,
            child_eventstream_index=copied_eventstream._eventstream_index,
        )
        return copied_eventstream

    @time_performance(
        scope="eventstream",
        event_name="append_eventstream",
    )
    def append_eventstream(self, eventstream: Eventstream) -> None:  # type: ignore
        """
        Append ``eventstream`` with the same schema.

        Parameters
        ----------
        eventstream : Eventstream

        Returns
        -------
        eventstream

        Raises
        ------
        ValueError
            If ``EventstreamSchemas`` of two ``eventstreams`` are not equal.
        """
        if not self.schema.is_equal(eventstream.schema):
            raise ValueError("invalid schema: joined eventstream")

        collect_data_performance(
            scope="eventstream",
            event_name="metadata",
            called_params={},
            performance_data={},
            eventstream_index=self._eventstream_index,
            parent_eventstream_index=eventstream._eventstream_index,
            child_eventstream_index=self._eventstream_index,
        )

        curr_events = self.to_dataframe()
        new_events = eventstream.to_dataframe()

        merged_events = pd.merge(
            curr_events,
            new_events,
            left_on=self.schema.event_id,
            right_on=self.schema.event_id,
            how="outer",
            indicator=True,
        )

        left_events = merged_events[(merged_events["_merge"] == "left_only")]
        both_events = merged_events[(merged_events["_merge"] == "both")]
        right_events = merged_events[(merged_events["_merge"] == "right_only")]

        right_events = pd.concat([right_events, both_events])

        cols = self.schema.get_cols()

        result_left_part = pd.DataFrame()
        result_right_part = pd.DataFrame()
        result_both_part = pd.DataFrame()

        with warnings.catch_warnings():
            # disable warning for pydantic schema Callable type
            warnings.simplefilter(action="ignore", category=FutureWarning)

            for col in cols:
                result_left_part[col] = get_merged_col(df=left_events, colname=col, suffix="_x")
                result_right_part[col] = get_merged_col(df=right_events, colname=col, suffix="_y")

        self.__events = pd.concat([result_left_part, result_both_part, result_right_part])
        self.index_events()

    def _get_both_custom_cols(self, eventstream: Eventstream) -> list[str]:
        self_custom_cols = set(self.schema.custom_cols)
        eventstream_custom_cols = set(eventstream.schema.custom_cols)
        all_custom_cols = self_custom_cols.union(eventstream_custom_cols)
        return list(all_custom_cols)

    def _get_both_cols(self, eventstream: Eventstream) -> list[str]:
        self_cols = set(self.schema.get_cols())
        eventstream_cols = set(eventstream.schema.get_cols())
        all_cols = self_cols.union(eventstream_cols)
        return list(all_cols)

    def _create_index(self, events: pd.DataFrame) -> pd.DataFrame:
        events_order_sort_col = "events_order_sort_col"
        events_type_sort_col = "events_type_sort_col"

        events[events_order_sort_col] = events[self.schema.event_name].apply(
            lambda e: self.__get_events_priority_by_config(e)
        )
        events[events_type_sort_col] = events[self.schema.event_type].apply(
            lambda e: self.__get_events_priority_by_type(e)
        )

        events = events.sort_values([self.schema.event_timestamp, events_order_sort_col, events_type_sort_col])  # type: ignore
        events = events.drop([events_order_sort_col, events_type_sort_col], axis=1)
        events.reset_index(inplace=True, drop=True)
        events[self.schema.event_index] = events.index
        return events

    @time_performance(
        scope="eventstream",
        event_name="to_dataframe",
    )
    def to_dataframe(self, copy: bool = False) -> pd.DataFrame:
        """
        Convert ``eventstream`` to ``pd.Dataframe``

        Parameters
        ----------
        copy : bool, default False
            If ``True`` - copy data from current ``eventstream``.
            See details in the :pandas_copy:`pandas documentation<>`.

        Returns
        -------
        pd.DataFrame

        """
        params: dict[str, Any] = {
            "copy": copy,
        }

        events = self.__events
        view = pd.DataFrame(events, columns=self.schema.get_cols(), copy=copy)
        self.__track_dataset(name="metadata", data=view, params=params, schema=self.schema, not_hash_values=[])
        return view

    @time_performance(
        scope="eventstream",
        event_name="index_events",
    )
    def index_events(self) -> None:
        """
        Sort and index eventstream using DEFAULT_INDEX_ORDER.

        Returns
        -------
        None

        """
        collect_data_performance(
            scope="eventstream",
            event_name="metadata",
            called_params={},
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        order_temp_col_name = "order"
        indexed = self.__events

        indexed[order_temp_col_name] = indexed[self.schema.event_type].apply(
            lambda e: self.__get_events_priority_by_type(e)
        )
        indexed = indexed.sort_values([self.schema.event_timestamp, self.schema.event_index, order_temp_col_name])  # type: ignore
        indexed = indexed.drop([order_temp_col_name], axis=1)
        indexed.reset_index(inplace=True, drop=True)
        self.__events = indexed

    @time_performance(
        scope="eventstream",
        event_name="add_custom_col",
    )
    def add_custom_col(self, name: str, data: pd.Series[Any] | None) -> None:
        """
        Add custom column to an existing ``eventstream``.

        Parameters
        ----------
        name : str
            New column name.
        data : pd.Series

            - If ``pd.Series`` - new column with given values will be added.
            - If ``None`` - new column will be filled with ``np.nan``.

        Returns
        -------
        Eventstream
        """
        collect_data_performance(
            scope="eventstream",
            event_name="metadata",
            called_params={"name": name, "data": data},
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__raw_data_schema.custom_cols.extend([{"custom_col": name, "raw_data_col": name}])
        self.schema.custom_cols.extend([name])
        self.__events[name] = data

    def __define_default_custom_cols(self, raw_data: pd.DataFrame | pd.Series[Any]) -> List[str]:
        raw_data_cols = self.__raw_data_schema.get_default_cols()
        schema_cols = self.schema.get_default_cols()

        cols_denylist: List[str] = raw_data_cols + schema_cols

        custom_cols: List[str] = []

        for raw_col_name in raw_data.columns:
            if raw_col_name in cols_denylist:
                continue
            custom_cols.append(raw_col_name)

        return custom_cols

    def __required_cleanup(self, events: pd.DataFrame | pd.Series[Any]) -> pd.DataFrame | pd.Series[Any]:
        income_size = len(events)
        events.dropna(  # type: ignore
            subset=[self.schema.event_name, self.schema.event_timestamp, self.schema.user_id], inplace=True
        )
        size_after_cleanup = len(events)
        if (removed_rows := income_size - size_after_cleanup) > 0:
            warnings.warn(
                "Removed %s rows because they have empty %s or %s or %s"
                % (removed_rows, self.schema.event_name, self.schema.event_timestamp, self.schema.user_id)
            )
        return events

    def __prepare_events(self, raw_data: pd.DataFrame | pd.Series[Any]) -> pd.DataFrame | pd.Series[Any]:
        events = raw_data.copy()

        # clear df
        events.drop(list(events.columns), axis=1, inplace=True)

        if self.__raw_data_schema.event_id is not None and self.__raw_data_schema.event_id in raw_data.columns:
            events[self.schema.event_id] = raw_data[self.__raw_data_schema.event_id]
        else:
            events[self.schema.event_id] = [uuid.uuid4() for x in range(len(events))]

        events[self.schema.event_name] = self.__get_col_from_raw_data(
            raw_data=raw_data,
            colname=self.__raw_data_schema.event_name,
        )
        events[self.schema.event_timestamp] = pd.to_datetime(
            self.__get_col_from_raw_data(
                raw_data=raw_data,
                colname=self.__raw_data_schema.event_timestamp,
            ),
        )
        events[self.schema.user_id] = self.__get_col_from_raw_data(
            raw_data=raw_data,
            colname=self.__raw_data_schema.user_id,
        )

        if self.__raw_data_schema.event_type is not None:
            events[self.schema.event_type] = self.__get_col_from_raw_data(
                raw_data=raw_data,
                colname=self.__raw_data_schema.event_type,
            )
        else:
            events[self.schema.event_type] = "raw"

        for custom_col_schema in self.__raw_data_schema.custom_cols:
            raw_data_col = custom_col_schema["raw_data_col"]
            custom_col = custom_col_schema["custom_col"]
            if custom_col not in self.schema.custom_cols:
                self.schema.custom_cols.append(custom_col)

            events[custom_col] = self.__get_col_from_raw_data(
                raw_data=raw_data,
                colname=raw_data_col,
            )

        for custom_col in self.schema.custom_cols:
            if custom_col in events.columns:
                continue
            events[custom_col] = np.nan

        if self.__raw_data_schema.event_index is not None and self.__raw_data_schema.event_index in raw_data.columns:
            events[self.schema.event_index] = raw_data[self.__raw_data_schema.event_index].astype("int64")
        else:
            events = self._create_index(events=events)  # type: ignore

        return events

    def __apply_default_dataprocessors(self, add_start_end_events: bool) -> None:
        from outhad_edge.data_processors_lib.add_start_end_events import (
            AddStartEndEvents,
            AddStartEndEventsParams,
        )

        events = self.__events
        name_col = self.schema.event_name

        if add_start_end_events:
            add_start_end_processor = AddStartEndEvents(AddStartEndEventsParams())
            self.__events = add_start_end_processor.apply(self.__events, self.schema)  # type: ignore

    def __get_col_from_raw_data(
        self, raw_data: pd.DataFrame | pd.Series[Any], colname: str, create: bool = False
    ) -> pd.Series | float:
        if colname in raw_data.columns:
            return raw_data[colname]
        else:
            if create:
                return np.nan
            else:
                raise ValueError(f'invalid raw data. Column "{colname}" does not exists!')

    def __get_events_priority_by_type(self, event_type: Optional[str]) -> int:
        if event_type in self.index_order:
            return self.index_order.index(event_type)
        return len(self.index_order)

    def __get_events_priority_by_config(self, event_name: str) -> int:
        if event_name in self.events_order:
            return self.events_order.index(event_name)
        return len(self.events_order)

    def __sample_user_paths(
        self,
        raw_data: pd.DataFrame | pd.Series[Any],
        raw_data_schema: RawDataSchemaType,
        user_sample_size: Optional[int | float] = None,
        user_sample_seed: Optional[int] = None,
    ) -> pd.DataFrame | pd.Series[Any]:
        if type(user_sample_size) is not float and type(user_sample_size) is not int:
            raise TypeError('"user_sample_size" has to be a number(float for user share or int for user amount)')
        if user_sample_size < 0:
            raise ValueError("User sample size/share cannot be negative!")
        if type(user_sample_size) is float:
            if user_sample_size > 1:
                raise ValueError("User sample share cannot exceed 1!")
        user_col_name = raw_data_schema.user_id
        unique_users = raw_data[user_col_name].unique()
        if type(user_sample_size) is int:
            sample_size = user_sample_size
        elif type(user_sample_size) is float:
            sample_size = int(user_sample_size * len(unique_users))
        else:
            return raw_data
        if user_sample_seed is not None:
            np.random.seed(user_sample_seed)
        sample_users = np.random.choice(unique_users, sample_size, replace=False)
        raw_data_sampled = raw_data.loc[raw_data[user_col_name].isin(sample_users), :]  # type: ignore
        return raw_data_sampled

    @time_performance(
        scope="funnel",
        event_name="helper",
        event_value="plot",
    )
    def funnel(
        self,
        stages: list[str],
        stage_names: list[str] | None = None,
        funnel_type: Literal["open", "closed", "hybrid"] = "closed",
        segments: Collection[Collection[int]] | None = None,
        segment_names: list[str] | None = None,
        show_plot: bool = True,
    ) -> Funnel:
        """
        Show a visualization of the user sequential events represented as a funnel.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, a funnel visualization is shown.
        See other parameters' description
            :py:class:`.Funnel`

        Returns
        -------
        Funnel
            A ``Funnel`` class instance fitted to the given parameters.

        """
        params = {
            "stages": stages,
            "stage_names": stage_names,
            "funnel_type": funnel_type,
            "segments": segments,
            "segment_names": segment_names,
            "show_plot": show_plot,
        }

        collect_data_performance(
            scope="funnel",
            event_name="metadata",
            called_params=params,
            not_hash_values=["funnel_type"],
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        self.__funnel = Funnel(eventstream=self)
        self.__funnel.fit(
            stages=stages,
            stage_names=stage_names,
            funnel_type=funnel_type,
            segments=segments,
            segment_names=segment_names,
        )
        if show_plot:
            figure = self.__funnel.plot()
            figure.show()
        return self.__funnel

    @property
    @time_performance(
        scope="clusters",
        event_name="helper",
        event_value="init",
    )
    def clusters(self) -> Clusters:
        """

        Returns
        -------
        Clusters
            A blank (not fitted) instance of ``Clusters`` class to be used for cluster analysis.

        See Also
        --------
        .Clusters
        """
        if self.__clusters is None:
            self.__clusters = Clusters(eventstream=self)
        return self.__clusters

    @time_performance(
        scope="step_matrix",
        event_name="helper",
        event_value="plot",
    )
    def step_matrix(
        self,
        max_steps: int = 20,
        weight_col: str | None = None,
        precision: int = 2,
        targets: list[str] | str | None = None,
        accumulated: Literal["both", "only"] | None = None,
        sorting: list | None = None,
        threshold: float = 0.01,
        centered: dict | None = None,
        groups: Tuple[list, list] | None = None,
        show_plot: bool = True,
    ) -> StepMatrix:
        """
        Show a heatmap visualization of the step matrix.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, a step matrix heatmap is shown.
        See other parameters' description
            :py:class:`.StepMatrix`

        Returns
        -------
        StepMatrix
            A ``StepMatrix`` class instance fitted to the given parameters.

        """

        params = {
            "max_steps": max_steps,
            "weight_col": weight_col,
            "precision": precision,
            "targets": targets,
            "accumulated": accumulated,
            "sorting": sorting,
            "threshold": threshold,
            "centered": centered,
            "groups": groups,
            "show_plot": show_plot,
        }
        not_hash_values = ["accumulated", "centered"]
        collect_data_performance(
            scope="step_matrix",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        self.__step_matrix = StepMatrix(eventstream=self)

        self.__step_matrix.fit(
            max_steps=max_steps,
            weight_col=weight_col,
            precision=precision,
            targets=targets,
            accumulated=accumulated,
            sorting=sorting,
            threshold=threshold,
            centered=centered,
            groups=groups,
        )
        if show_plot:
            self.__step_matrix.plot()
        return self.__step_matrix

    @time_performance(
        scope="step_sankey",
        event_name="helper",
        event_value="plot",
    )
    def step_sankey(
        self,
        max_steps: int = 10,
        threshold: int | float = 0.05,
        sorting: list | None = None,
        targets: list[str] | str | None = None,
        autosize: bool = True,
        width: int | None = None,
        height: int | None = None,
        show_plot: bool = True,
    ) -> StepSankey:
        """
        Show a Sankey diagram visualizing the user paths in stepwise manner.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, a sankey diagram is shown.
        See other parameters' description
            :py:class:`.StepSankey`

        Returns
        -------
        StepSankey
            A ``StepSankey`` class instance fitted to the given parameters.

        """

        params = {
            "max_steps": max_steps,
            "threshold": threshold,
            "sorting": sorting,
            "targets": targets,
            "autosize": autosize,
            "width": width,
            "height": height,
            "show_plot": show_plot,
        }

        collect_data_performance(
            scope="step_sankey",
            event_name="metadata",
            called_params=params,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__sankey = StepSankey(eventstream=self)

        self.__sankey.fit(max_steps=max_steps, threshold=threshold, sorting=sorting, targets=targets)
        if show_plot:
            figure = self.__sankey.plot(autosize=autosize, width=width, height=height)
            figure.show()
        return self.__sankey

    @time_performance(
        scope="cohorts",
        event_name="helper",
        event_value="heatmap",
    )
    def cohorts(
        self,
        cohort_start_unit: DATETIME_UNITS,
        cohort_period: Tuple[int, DATETIME_UNITS],
        average: bool = True,
        cut_bottom: int = 0,
        cut_right: int = 0,
        cut_diagonal: int = 0,
        width: float = 5.0,
        height: float = 5.0,
        show_plot: bool = True,
    ) -> Cohorts:
        """
        Show a heatmap visualization of the user appearance grouped by cohorts.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, a cohort matrix heatmap is shown.
        See other parameters' description
            :py:class:`.Cohorts`

        Returns
        -------
        Cohorts
            A ``Cohorts`` class instance fitted to the given parameters.
        """

        params = {
            "cohort_start_unit": cohort_start_unit,
            "cohort_period": cohort_period,
            "average": average,
            "cut_bottom": cut_bottom,
            "cut_right": cut_right,
            "cut_diagonal": cut_diagonal,
            "width": width,
            "height": height,
            "show_plot": show_plot,
        }

        not_hash_values = ["cohort_start_unit", "cohort_period"]
        collect_data_performance(
            scope="cohorts",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__cohorts = Cohorts(eventstream=self)

        self.__cohorts.fit(
            cohort_start_unit=cohort_start_unit,
            cohort_period=cohort_period,
            average=average,
            cut_bottom=cut_bottom,
            cut_right=cut_right,
            cut_diagonal=cut_diagonal,
        )
        if show_plot:
            self.__cohorts.heatmap(width=width, height=height)
        return self.__cohorts

    @time_performance(
        scope="nlq",
        event_name="helper",
        event_value="init",
    )
    def nlq(
        self,
        model: str = "gpt-4",
        temperature: float = 0.1,
        max_retries: int = 3,
        use_cache: bool = True,
        verbose: bool = False,
    ):
        """
        Create a Natural Language Query interface for the eventstream.

        Enables asking questions about your data in natural language using LLMs (GPT-4, Claude).
        The system uses RAG (Retrieval Augmented Generation), semantic caching, and sandboxed
        code execution to provide accurate answers.

        Parameters
        ----------
        model : str, default "gpt-4"
            LLM model to use. Options:
            - "gpt-4" (recommended for accuracy)
            - "gpt-3.5-turbo" (faster, cheaper)
            - "claude-3-sonnet" (Anthropic)
            - "claude-3-opus" (most capable)
        temperature : float, default 0.1
            LLM temperature (0.0-1.0). Lower = more deterministic.
        max_retries : int, default 3
            Maximum retry attempts for self-correction.
        use_cache : bool, default True
            Enable semantic caching for faster repeated queries.
        verbose : bool, default False
            Print debug information during query execution.

        Returns
        -------
        NaturalLanguageQuery
            An NLQ instance ready to answer questions via `.ask(question)`.

        Examples
        --------
        >>> stream = Eventstream(df)
        >>> nlq = stream.nlq(model="gpt-4")
        >>> result = nlq.ask("What is the conversion rate?")
        >>> print(result['answer'])
        'The conversion rate is 3.5%'

        >>> # With custom model
        >>> nlq = stream.nlq(model="claude-3-sonnet", temperature=0.2)
        >>> result = nlq.ask("Show me the most common user paths")

        See Also
        --------
        :py:class:`.NaturalLanguageQuery`
            Full documentation of the NLQ system.

        Notes
        -----
        Requires AI dependencies: ``pip install outhad_edge[ai]``

        Set up API keys in .env file:
        - OPENAI_API_KEY=sk-...
        - ANTHROPIC_API_KEY=sk-ant-...
        """
        try:
            from outhad_edge.ai import NaturalLanguageQuery, check_ai_dependencies
        except ImportError:
            raise ImportError(
                "Natural Language Query requires AI dependencies.\n\n"
                "Install them with:\n"
                "  pip install outhad_edge[ai]\n\n"
                "Or manually:\n"
                "  pip install -r requirements-ai.txt\n\n"
                "Then set up your API keys in .env file."
            )

        if not check_ai_dependencies():
            raise ImportError(
                "AI dependencies are not fully installed. "
                "Run: pip install outhad_edge[ai]"
            )

        params = {
            "model": model,
            "temperature": temperature,
            "max_retries": max_retries,
            "use_cache": use_cache,
        }

        collect_data_performance(
            scope="nlq",
            event_name="metadata",
            called_params=params,
            not_hash_values=["model", "use_cache"],
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        # Create NLQ instance
        nlq_instance = NaturalLanguageQuery(
            eventstream=self,
            model=model,
            temperature=temperature,
            max_retries=max_retries,
            use_cache=use_cache,
            verbose=verbose,
        )

        return nlq_instance

    @time_performance(
        scope="stattests",
        event_name="helper",
        event_value="display_results",
    )
    def stattests(
        self,
        test: STATTEST_NAMES,
        groups: Tuple[list[str | int], list[str | int]],
        func: Callable,
        group_names: Tuple[str, str],
        alpha: float = 0.05,
    ) -> StatTests:
        """
        Determine the statistical difference between the metric values in two user groups.

        Parameters
        ----------
        See parameters' description
            :py:class:`.Stattests`

        Returns
        -------
        StatTests
            A ``StatTest`` class instance fitted to the given parameters.
        """
        params = {
            "test": test,
            "groups": groups,
            "func": func,
            "group_names": group_names,
            "alpha": alpha,
        }
        not_hash_values = ["test"]

        collect_data_performance(
            scope="stattests",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__stattests = StatTests(eventstream=self)
        self.__stattests.fit(groups=groups, func=func, test=test, group_names=group_names, alpha=alpha)
        self.__stattests.display_results()
        return self.__stattests

    @time_performance(
        scope="timedelta_hist",
        event_name="helper",
        event_value="plot",
    )
    def timedelta_hist(
        self,
        raw_events_only: bool = False,
        event_pair: list[str | Literal[EVENTSTREAM_GLOBAL_EVENTS]] | None = None,
        adjacent_events_only: bool = True,
        weight_col: str | None = None,
        time_agg: AGGREGATION_NAMES | None = None,
        timedelta_unit: DATETIME_UNITS = "s",
        log_scale: bool | tuple[bool, bool] | None = None,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
        bins: int | Literal[BINS_ESTIMATORS] = 20,
        width: float = 6.0,
        height: float = 4.5,
        show_plot: bool = True,
    ) -> TimedeltaHist:
        """
        Plot the distribution of the time deltas between two events. Support various
        distribution types, such as distribution of time for adjacent consecutive events, or
        for a pair of pre-defined events, or median transition time from event to event per user/session.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, histogram is shown.
        See other parameters' description
            :py:class:`.TimedeltaHist`

        Returns
        -------
        TimedeltaHist
            A ``TimedeltaHist`` class instance fitted with given parameters.

        """

        params = {
            "raw_events_only": raw_events_only,
            "event_pair": event_pair,
            "adjacent_events_only": adjacent_events_only,
            "weight_col": weight_col,
            "time_agg": time_agg,
            "timedelta_unit": timedelta_unit,
            "log_scale": log_scale,
            "lower_cutoff_quantile": lower_cutoff_quantile,
            "upper_cutoff_quantile": upper_cutoff_quantile,
            "bins": bins,
            "width": width,
            "height": height,
            "show_plot": show_plot,
        }
        not_hash_values = ["time_agg", "timedelta_unit"]

        collect_data_performance(
            scope="timedelta_hist",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__timedelta_hist = TimedeltaHist(
            eventstream=self,
        )

        self.__timedelta_hist.fit(
            raw_events_only=raw_events_only,
            event_pair=event_pair,
            adjacent_events_only=adjacent_events_only,
            time_agg=time_agg,
            weight_col=weight_col,
            timedelta_unit=timedelta_unit,
            log_scale=log_scale,
            lower_cutoff_quantile=lower_cutoff_quantile,
            upper_cutoff_quantile=upper_cutoff_quantile,
            bins=bins,
        )
        if show_plot:
            self.__timedelta_hist.plot(
                width=width,
                height=height,
            )

        return self.__timedelta_hist

    @time_performance(
        scope="user_lifetime_hist",
        event_name="helper",
        event_value="plot",
    )
    def user_lifetime_hist(
        self,
        timedelta_unit: DATETIME_UNITS = "s",
        log_scale: bool | tuple[bool, bool] | None = None,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
        bins: int | Literal[BINS_ESTIMATORS] = 20,
        width: float = 6.0,
        height: float = 4.5,
        show_plot: bool = True,
    ) -> UserLifetimeHist:
        """
        Plot the distribution of user lifetimes. A ``users lifetime`` is the timedelta between the first and the last
        events of the user.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, histogram is shown.
        See other parameters' description
            :py:class:`.UserLifetimeHist`

        Returns
        -------
        UserLifetimeHist
            A ``UserLifetimeHist`` class instance with given parameters.


        """
        params = {
            "timedelta_unit": timedelta_unit,
            "log_scale": log_scale,
            "lower_cutoff_quantile": lower_cutoff_quantile,
            "upper_cutoff_quantile": upper_cutoff_quantile,
            "bins": bins,
            "width": width,
            "height": height,
            "show_plot": show_plot,
        }
        not_hash_values = ["timedelta_unit"]

        collect_data_performance(
            scope="user_lifetime_hist",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        self.__user_lifetime_hist = UserLifetimeHist(
            eventstream=self,
        )
        self.__user_lifetime_hist.fit(
            timedelta_unit=timedelta_unit,
            log_scale=log_scale,
            lower_cutoff_quantile=lower_cutoff_quantile,
            upper_cutoff_quantile=upper_cutoff_quantile,
            bins=bins,
        )
        if show_plot:
            self.__user_lifetime_hist.plot(width=width, height=height)
        return self.__user_lifetime_hist

    @time_performance(
        scope="event_timestamp_hist",
        event_name="helper",
        event_value="plot",
    )
    def event_timestamp_hist(
        self,
        event_list: list[str] | None = None,
        raw_events_only: bool = False,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
        bins: int | Literal[BINS_ESTIMATORS] = 20,
        width: float = 6.0,
        height: float = 4.5,
        show_plot: bool = True,
    ) -> EventTimestampHist:
        """
        Plot distribution of events over time. Can be useful for detecting time-based anomalies, and visualising
        general timespan of the eventstream.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, histogram is shown.
        See other parameters' description
            :py:class:`.EventTimestampHist`


        Returns
        -------
        EventTimestampHist
            A ``EventTimestampHist`` class instance with given parameters.
        """
        params = {
            "event_list": event_list,
            "raw_events_only": raw_events_only,
            "lower_cutoff_quantile": lower_cutoff_quantile,
            "upper_cutoff_quantile": upper_cutoff_quantile,
            "bins": bins,
            "width": width,
            "height": height,
            "show_plot": show_plot,
        }

        collect_data_performance(
            scope="event_timestamp_hist",
            event_name="metadata",
            called_params=params,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        self.__event_timestamp_hist = EventTimestampHist(
            eventstream=self,
        )

        self.__event_timestamp_hist.fit(
            event_list=event_list,
            raw_events_only=raw_events_only,
            lower_cutoff_quantile=lower_cutoff_quantile,
            upper_cutoff_quantile=upper_cutoff_quantile,
            bins=bins,
        )
        if show_plot:
            self.__event_timestamp_hist.plot(width=width, height=height)
        return self.__event_timestamp_hist

    @time_performance(
        scope="describe",
        event_name="helper",
        event_value="_values",
    )
    def describe(self, session_col: str = "session_id", raw_events_only: bool = False) -> pd.DataFrame:
        """
        Display general eventstream information. If ``session_col`` is present in eventstream, also
        output session statistics.

        Parameters
        ----------
        session_col : str, default 'session_id'
            Specify name of the session column. If the column is present in the eventstream,
            session statistics will be added to the output.

        raw_events_only : bool, default False
            If ``True`` - statistics will only be shown for raw events.
            If ``False`` - statistics will be shown for all events presented in your data.

        Returns
        -------
        pd.DataFrame
            A dataframe containing descriptive statistics for the eventstream.


        See Also
        --------
        .EventTimestampHist : Plot the distribution of events over time.
        .TimedeltaHist : Plot the distribution of the time deltas between two events.
        .UserLifetimeHist : Plot the distribution of user lifetimes.
        .Eventstream.describe_events : Show general eventstream events statistics.


        Notes
        -----
        - All ``float`` values are rounded to 2.
        - All ``datetime`` values are rounded to seconds.

        See :ref:`Eventstream user guide<eventstream_describe>` for the details.


        """
        params = {
            "session_col": session_col,
            "raw_events_only": raw_events_only,
        }

        collect_data_performance(
            scope="describe",
            event_name="metadata",
            called_params=params,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )
        describer = _Describe(eventstream=self, session_col=session_col, raw_events_only=raw_events_only)
        return describer._values()

    @time_performance(
        scope="describe_events",
        event_name="helper",
        event_value="_values",
    )
    def describe_events(
        self, session_col: str = "session_id", raw_events_only: bool = False, event_list: list[str] | None = None
    ) -> pd.DataFrame:
        """
        Display general information on eventstream events. If ``session_col`` is present in eventstream, also
        output session statistics.

        Parameters
        ----------
        session_col : str, default 'session_id'
            Specify name of the session column. If the column is present in the eventstream,
            output session statistics.

        raw_events_only : bool, default False
            If ``True`` - statistics will only be shown for raw events.
            If ``False`` - statistics will be shown for all events presented in your data.

        event_list : list of str, optional
            Specify events to be displayed.

        Returns
        -------
        pd.DataFrame
            **Eventstream statistics**:

            - The following metrics are calculated for each event present in the eventstream
              (or the narrowed eventstream if parameters ``event_list`` or ``raw_events_only`` are used).
              Let all_events, all_users, all_sessions be the numbers of all events, users,
              and sessions present in the eventstream. Then:

                - *number_of_occurrences* - the number of occurrences of a particular event in the eventstream;
                - *unique_users* - the number of unique users who experienced a particular event;
                - *unique_sessions* - the number of unique sessions with each event;
                - *number_of_occurrences_shared* - number_of_occurrences / all_events (raw_events_only,
                  if this parameter = ``True``);
                - *unique_users_shared* - unique_users / all_users;
                - *unique_sessions_shared* - unique_sessions / all_sessions;

            - **time_to_FO_user_wise** category - timedelta between ``path_start``
              and the first occurrence (FO) of a specified event in each user path.
            - **steps_to_FO_user_wise** category - the number of steps (events) from
              ``path_start`` to the first occurrence (FO) of a specified event in each user path.
              If ``raw_events_only=True`` only raw events will be counted.
            - **time_to_FO_session_wise** category - timedelta  between ``session_start``
              and the first occurrence (FO) of a specified event in each session.
            - **steps_to_FO_session_wise** category - the number of steps (events) from
              ``session_start`` to the first occurrence (FO) of a specified event in each session.
              If ``raw_events_only=True`` only raw events will be counted.

            Agg functions for each ``first_occurrence*`` category are: mean, std, median, min, max.

        See Also
        --------
        .EventTimestampHist : Plot the distribution of events over time.
        .TimedeltaHist : Plot the distribution of the time deltas between two events.
        .UserLifetimeHist : Plot the distribution of user lifetimes.
        .Eventstream.describe : Show general eventstream statistics.

        Notes
        -----
        - All ``float`` values are rounded to 2.
        - All ``datetime`` values are rounded to seconds.

        See :ref:`Eventstream user guide<eventstream_describe_events>` for the details.

        """

        params = {
            "session_col": session_col,
            "raw_events_only": raw_events_only,
            "event_list": event_list,
        }

        collect_data_performance(
            scope="describe_events",
            event_name="metadata",
            called_params=params,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        describer = _DescribeEvents(
            eventstream=self, session_col=session_col, event_list=event_list, raw_events_only=raw_events_only
        )
        return describer._values()

    @time_performance(
        scope="transition_graph",
        event_name="helper",
        event_value="plot",
    )
    def transition_graph(
        self,
        edges_norm_type: NormType = None,
        nodes_norm_type: NormType = None,
        targets: MutableMapping[str, str | None] | None = None,
        nodes_threshold: Threshold | None = None,
        edges_threshold: Threshold | None = None,
        nodes_weight_col: str | None = None,
        edges_weight_col: str | None = None,
        custom_weight_cols: list[str] | None = None,
        width: str | int | float = "100%",
        height: str | int | float = "60vh",
        show_weights: bool = True,
        show_percents: bool = False,
        show_nodes_names: bool = True,
        show_all_edges_for_targets: bool = True,
        show_nodes_without_links: bool = False,
        show_edge_info_on_hover: bool = True,
        layout_dump: str | None = None,
        nodes_custom_colors: Dict[str, str] | None = None,
        edges_custom_colors: Dict[Tuple[str, str], str] | None = None,
    ) -> TransitionGraph:
        """

        Parameters
        ----------
        See parameters' description
            :py:meth:`.TransitionGraph.plot`
            @TODO: maybe load docs with docrep? 2dpanina, Vladimir Makhanov

        Returns
        -------
        TransitionGraph
            Rendered IFrame graph.

        """

        params = {
            "edges_norm_type": edges_norm_type,
            "nodes_norm_type": nodes_norm_type,
            "targets": targets,
            "nodes_threshold": nodes_threshold,
            "edges_threshold": edges_threshold,
            "nodes_weight_col": nodes_weight_col,
            "edges_weight_col": edges_weight_col,
            "custom_weight_cols": custom_weight_cols,
            "width": width,
            "height": height,
            "show_weights": show_weights,
            "show_percents": show_percents,
            "show_nodes_names": show_nodes_names,
            "show_all_edges_for_targets": show_all_edges_for_targets,
            "show_nodes_without_links": show_nodes_without_links,
            "show_edge_info_on_hover": show_edge_info_on_hover,
            "layout_dump": layout_dump,
        }
        not_hash_values = ["edges_norm_type", "targets", "width", "height"]

        collect_data_performance(
            scope="transition_graph",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__transition_graph = TransitionGraph(eventstream=self)
        self.__transition_graph.plot(
            targets=targets,
            edges_norm_type=edges_norm_type,
            nodes_norm_type=nodes_norm_type,
            edges_weight_col=edges_weight_col,
            nodes_threshold=nodes_threshold,
            edges_threshold=edges_threshold,
            nodes_weight_col=nodes_weight_col,
            custom_weight_cols=custom_weight_cols,
            width=width,
            height=height,
            show_weights=show_weights,
            show_percents=show_percents,
            show_nodes_names=show_nodes_names,
            show_all_edges_for_targets=show_all_edges_for_targets,
            show_nodes_without_links=show_nodes_without_links,
            layout_dump=layout_dump,
            show_edge_info_on_hover=show_edge_info_on_hover,
            nodes_custom_colors=nodes_custom_colors,
            edges_custom_colors=edges_custom_colors,
        )
        return self.__transition_graph

    @time_performance(
        scope="preprocessing_graph",
        event_name="helper",
        event_value="display",
    )
    def preprocessing_graph(self, width: int = 960, height: int = 600) -> PreprocessingGraph:
        """
        Display the preprocessing GUI tool.

        Parameters
        ----------
        width : int, default 960
            Width of plot in pixels.
        height : int, default 600
            Height of plot in pixels.

        Returns
        -------
        PreprocessingGraph
            Rendered preprocessing graph.
        """

        params = {
            "width": width,
            "height": height,
        }

        collect_data_performance(
            scope="preprocessing_graph",
            event_name="metadata",
            called_params=params,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        if self._preprocessing_graph is None:
            self._preprocessing_graph = PreprocessingGraph(source_stream=self)
        self._preprocessing_graph.display(width=width, height=height)

        return self._preprocessing_graph

    @time_performance(
        scope="transition_matrix",
        event_name="helper",
        event_value="_values",
    )
    def transition_matrix(self, weight_col: str | None = None, norm_type: NormType = None) -> pd.DataFrame:
        """
        Get transition weights as a matrix for each unique pair of events. The calculation logic is the same
        that is used for edge weights calculation of transition graph.

        Parameters
        ----------

        weight_col : str, optional
            Weighting column for the transition weights calculation.
            See :ref:`transition graph user guide <transition_graph_weights>` for the details.

        norm_type : {"full", "node", None}, default None
            Normalization type. See :ref:`transition graph user guide <transition_graph_weights>` for the details.

        Returns
        -------
        pd.DataFrame
            Transition matrix. ``(i, j)``-th matrix value relates to the weight of i → j transition.

        Notes
        -----
        See :ref:`transition graph user guide <transition_graph_transition_matrix>` for the details.
        """
        not_hash_values = ["norm_type"]

        params = {"weight_col": weight_col, "norm_type": norm_type}
        collect_data_performance(
            scope="transition_matrix",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        matrix = _TransitionMatrix(eventstream=self)
        return matrix._values(weight_col=weight_col, norm_type=norm_type)

    @time_performance(
        scope="sequences",
        event_name="helper",
        event_value="plot",
    )
    def sequences(
        self,
        ngram_range: Tuple[int, int] = (1, 1),
        groups: Tuple[list, list] | None = None,
        group_names: Tuple[str, str] | None = None,
        weight_col: str | None = None,
        metrics: MetricsType | None = None,
        threshold: tuple[str, float | int] | None = None,
        sorting: tuple[str | tuple, bool] | tuple[list[str | tuple], list[bool]] | None = None,
        heatmap_cols: str | list[str | tuple] | None = None,
        sample_size: int | None = 1,
        precision: int = 2,
        show_plot: bool = True,
    ) -> Sequences:
        """
        Calculate statistics on n-grams found in eventstream.

        Parameters
        ----------
        show_plot : bool, default True
            If ``True``, a sankey diagram is shown.
        See other parameters' description
            :py:class:`.Sequences`

        Returns
        -------
        Sequences
            A ``Sequences`` class instance fitted to the given parameters.

        """

        params = {
            "ngram_range": ngram_range,
            "groups": groups,
            "group_names": group_names,
            "weight_col": weight_col,
            "metrics": metrics,
            "threshold": threshold,
            "sorting": sorting,
            "heatmap_cols": heatmap_cols,
            "sample_size": sample_size,
            "precision": precision,
            "show_plot": show_plot,
        }
        not_hash_values = ["metrics", "ngram_range"]

        collect_data_performance(
            scope="sequences",
            event_name="metadata",
            called_params=params,
            not_hash_values=not_hash_values,
            performance_data={},
            eventstream_index=self._eventstream_index,
        )

        self.__sequences = Sequences(eventstream=self)

        self.__sequences.fit(ngram_range=ngram_range, groups=groups, group_names=group_names, weight_col=weight_col)

        styler_sequences = self.__sequences.plot(
            metrics=metrics,
            threshold=threshold,
            sorting=sorting,
            heatmap_cols=heatmap_cols,
            sample_size=sample_size,
            precision=precision,
        )
        if show_plot:
            display(styler_sequences)
        return self.__sequences



"""
NLQ System Initialization Helper.

Provides utilities for one-time initialization of the Natural Language Query system.
Populates vector database with metadata, examples, and business glossary.
"""

from typing import Tuple, Optional
from .semantic_layer import SemanticLayer
from .example_queries import ExampleQueryLibrary
from .vector_store import VectorStore


def initialize_nlq_system(
    eventstream=None,
    persist_directory: Optional[str] = None,
    force_reload: bool = False,
    verbose: bool = True
) -> Tuple[VectorStore, SemanticLayer]:
    """
    One-time initialization of NLQ system.
    Populates vector database with metadata.

    Args:
        eventstream: Optional Eventstream object for schema extraction
        persist_directory: Optional directory for ChromaDB persistence
        force_reload: If True, clear existing data and reload
        verbose: Print progress messages

    Returns:
        tuple: (VectorStore, SemanticLayer) instances

    Example:
        >>> from outhad_edge import datasets
        >>> stream = datasets.load_simple_shop()
        >>> vector_store, semantic_layer = initialize_nlq_system(stream)
    """

    if verbose:
        print("=" * 70)
        print("Initializing NLQ system...")
        print("=" * 70)

    # Create semantic layer
    if verbose:
        print("\n[1/5] Creating semantic layer...")

    if eventstream:
        semantic_layer = SemanticLayer(eventstream)
        if verbose:
            print(f"✓ Semantic layer created for eventstream")
    else:
        # Create empty semantic layer (will need eventstream later)
        semantic_layer = None
        if verbose:
            print("⚠ No eventstream provided, skipping schema metadata")

    # Create vector store
    if verbose:
        print("\n[2/5] Creating vector store...")

    vector_store = VectorStore(persist_directory=persist_directory)

    if force_reload:
        if verbose:
            print("⚠ Force reload enabled, clearing existing data...")
        vector_store.clear()

    if verbose:
        print(f"✓ Vector store created at: {vector_store.persist_directory}")

    # Add schema metadata (if eventstream provided)
    if semantic_layer:
        if verbose:
            print("\n[3/5] Adding schema metadata...")

        try:
            vector_store.add_schema_metadata(semantic_layer)
            if verbose:
                schema_count = vector_store.schema_collection.count() if vector_store.schema_collection else 0
                print(f"✓ Added schema metadata ({schema_count} documents)")
        except Exception as e:
            if verbose:
                print(f"✗ Failed to add schema metadata: {e}")
    else:
        if verbose:
            print("\n[3/5] Skipping schema metadata (no eventstream)")

    # Add example queries
    if verbose:
        print("\n[4/5] Adding example queries...")

    try:
        examples = ExampleQueryLibrary.get_examples()

        # Add to both collections for backward compatibility
        vector_store.add_examples(examples)  # Main collection (backward compat)

        # Format for new examples_collection
        formatted_examples = []
        for ex in examples:
            formatted_examples.append({
                'query': ex['query'],
                'code': ex['code'],
                'answer': ex.get('answer', ''),
                'description': ex.get('description', ''),
                'tags': ex.get('tags', [])
            })

        vector_store.add_example_queries(formatted_examples)  # New collection

        if verbose:
            print(f"✓ Added {len(examples)} example queries")
            if vector_store.examples_collection:
                print(f"  - Examples collection: {vector_store.examples_collection.count()} documents")
            print(f"  - Main collection: {vector_store.collection.count()} documents")

    except Exception as e:
        if verbose:
            print(f"✗ Failed to add examples: {e}")

    # Add business glossary
    if semantic_layer:
        if verbose:
            print("\n[5/5] Adding business glossary...")

        try:
            vector_store.add_business_terms(semantic_layer.business_glossary)
            if verbose:
                glossary_count = vector_store.glossary_collection.count() if vector_store.glossary_collection else 0
                print(f"✓ Added business glossary ({glossary_count} terms)")
        except Exception as e:
            if verbose:
                print(f"✗ Failed to add glossary: {e}")
    else:
        if verbose:
            print("\n[5/5] Skipping business glossary (no eventstream)")

    if verbose:
        print("\n" + "=" * 70)
        print("✓ NLQ system initialized successfully!")
        print("=" * 70)

        print("\nCollection Status:")
        print(f"  - Main collection: {vector_store.collection.count()} documents")
        if vector_store.schema_collection:
            print(f"  - Schema metadata: {vector_store.schema_collection.count()} documents")
        if vector_store.examples_collection:
            print(f"  - Example queries: {vector_store.examples_collection.count()} documents")
        if vector_store.glossary_collection:
            print(f"  - Business glossary: {vector_store.glossary_collection.count()} terms")
        if vector_store.cache_collection:
            print(f"  - Query cache: {vector_store.cache_collection.count()} cached queries")

        print("\nUsage:")
        print("  >>> from outhad_edge import datasets")
        print("  >>> stream = datasets.load_simple_shop()")
        print("  >>> nlq = stream.nlq(model='gpt-4')")
        print("  >>> result = nlq.ask('What is the conversion rate?')")

    return vector_store, semantic_layer


def reinitialize_vector_store(
    eventstream=None,
    persist_directory: Optional[str] = None,
    verbose: bool = True
) -> Tuple[VectorStore, Optional[SemanticLayer]]:
    """
    Reinitialize vector store by clearing and reloading all data.

    Args:
        eventstream: Optional Eventstream object
        persist_directory: Optional directory for ChromaDB persistence
        verbose: Print progress messages

    Returns:
        tuple: (VectorStore, SemanticLayer) instances
    """
    return initialize_nlq_system(
        eventstream=eventstream,
        persist_directory=persist_directory,
        force_reload=True,
        verbose=verbose
    )


def check_nlq_status(vector_store: VectorStore) -> dict:
    """
    Check the current status of NLQ system initialization.

    Args:
        vector_store: VectorStore instance to check

    Returns:
        dict: Status information about each collection
    """
    status = {
        'initialized': False,
        'collections': {}
    }

    # Check main collection
    status['collections']['main'] = {
        'exists': vector_store.collection is not None,
        'count': vector_store.collection.count() if vector_store.collection else 0
    }

    # Check schema collection
    status['collections']['schema'] = {
        'exists': vector_store.schema_collection is not None,
        'count': vector_store.schema_collection.count() if vector_store.schema_collection else 0
    }

    # Check examples collection
    status['collections']['examples'] = {
        'exists': vector_store.examples_collection is not None,
        'count': vector_store.examples_collection.count() if vector_store.examples_collection else 0
    }

    # Check glossary collection
    status['collections']['glossary'] = {
        'exists': vector_store.glossary_collection is not None,
        'count': vector_store.glossary_collection.count() if vector_store.glossary_collection else 0
    }

    # Check cache collection
    status['collections']['cache'] = {
        'exists': vector_store.cache_collection is not None,
        'count': vector_store.cache_collection.count() if vector_store.cache_collection else 0
    }

    # Determine if initialized
    examples_count = status['collections']['examples']['count']
    status['initialized'] = examples_count > 0

    return status



from __future__ import annotations

from typing import Any, Literal, Optional

import pandas as pd

from outhad_edge.backend.tracker import collect_data_performance, time_performance
from outhad_edge.data_processor import DataProcessor
from outhad_edge.eventstream.types import EventstreamSchemaType
from outhad_edge.params_model import ParamsModel
from outhad_edge.utils.doc_substitution import docstrings
from outhad_edge.utils.hash_object import hash_dataframe


class TruncatePathsParams(ParamsModel):
    """
    A class with parameters for :py:class:`.TruncatePath` class.
    """

    drop_before: Optional[str]
    drop_after: Optional[str]
    occurrence_before: Literal["first", "last"] = "first"
    occurrence_after: Literal["first", "last"] = "first"
    shift_before: int = 0
    shift_after: int = 0


@docstrings.get_sections(base="TruncatePath")  # type: ignore
class TruncatePaths(DataProcessor):
    """
    Remove events that will be deleted from each user's path
    based on the specified event and selected parameters.

    Parameters
    ----------
    drop_before : str, optional
        Event name before which part of the user's path is dropped. The specified event remains in the data.
    drop_after : str, optional
        Event name after which part of the user's path is dropped. The specified event remains in the data.
    occurrence_before : {"first", "last"}, default "first"
        This parameter is necessary when the specified event occurs more than once in one user's path.

        - when set to ``first``, the part of the user path before the first event occurrence is dropped;
        - when set to ``last``, the part of the user path before the last event occurrence is dropped;
    occurrence_after : {"first", "last"}, default "first"
        The same behavior as in the 'occurrence_before', but for the other part of the user path.
    shift_before : int,  default 0
        Sets the number of steps by which the truncate point is shifted from the selected event.
        If the value is negative, then the offset occurs to the left along the timeline.
        If positive, then it occurs to the right.
    shift_after : int,  default 0
        The same behavior as in the ``shift_before``, but for the other part of the user path.

    Returns
    -------
    Eventstream
        ``Eventstream`` with events that should be deleted from input ``eventstream``.


    Notes
    -----
    ``Step`` - is the group of events in the user path with the same timestamp.
    If the user path doesn't contain events from ``drop_before`` and ``drop_after`` parameters, then its
    path does not change.

    See :doc:`Data processors user guide</user_guides/dataprocessors>` for the details.
    """

    params: TruncatePathsParams

    @time_performance(
        scope="truncate_paths",
        event_name="init",
    )
    def __init__(self, params: TruncatePathsParams):
        super().__init__(params=params)

    @time_performance(
        scope="truncate_paths",
        event_name="apply",
    )
    def apply(self, df: pd.DataFrame, schema: EventstreamSchemaType) -> pd.DataFrame:
        source = df.copy()

        user_col = schema.user_id
        time_col = schema.event_timestamp
        event_col = schema.event_name
        event_id = schema.event_id

        drop_before = self.params.drop_before
        drop_after = self.params.drop_after
        occurrence_before = self.params.occurrence_before
        occurrence_after = self.params.occurrence_after
        shift_before = self.params.shift_before
        shift_after = self.params.shift_after

        params_data: list[Any] = []

        if not drop_after and not drop_before:
            raise Exception("Either drop_before or drop_after must be specified!")

        if drop_before:
            before: list[str | list[str | int | None]] | None = [
                drop_before,
                ["before", occurrence_before, shift_before],
            ]
            params_data.append(before)

        if drop_after:
            after: list[str | list[str | int | None]] | None = [drop_after, ["after", occurrence_after, shift_after]]
            params_data.append(after)

        for truncate_type in params_data:
            col_mark, occurrence, shift = truncate_type[1]

            if truncate_type[0]:
                mask_events = df[event_col] == truncate_type[0]
                df[f"{col_mark}_mark_target"] = mask_events.astype(int)
                df[f"{col_mark}_mark_target"] = df.groupby([user_col, time_col])[f"{col_mark}_mark_target"].transform(
                    max
                )
                if occurrence == "last":
                    df[f"{col_mark}_cumsum"] = df.iloc[::-1].groupby([user_col])[f"{col_mark}_mark_target"].cumsum()
                if occurrence == "first":
                    df[f"{col_mark}_cumsum"] = df.groupby([user_col])[f"{col_mark}_mark_target"].cumsum()

                def count_groups(x: pd.DataFrame) -> int:
                    return x.to_frame(name=time_col).groupby(time_col).ngroup()  # type: ignore

                df[f"{col_mark}_group_num_in_user"] = df.groupby([user_col], group_keys=False)[time_col].transform(
                    count_groups
                )

                if occurrence == "last":
                    df_groups = (
                        df[df[f"{col_mark}_cumsum"] == 1]
                        .groupby([user_col])[f"{col_mark}_group_num_in_user"]
                        .max()
                        .rename(f"{col_mark}_group_centered")
                        .reset_index()
                    )
                else:
                    df_groups = (
                        df[df[f"{col_mark}_cumsum"] == 1]
                        .groupby([user_col])[f"{col_mark}_group_num_in_user"]
                        .min()
                        .rename(f"{col_mark}_group_centered")
                        .reset_index()
                    )

                df = df.merge(df_groups)
                df[f"{col_mark}_group_centered"] = (
                    df[f"{col_mark}_group_num_in_user"] - df[f"{col_mark}_group_centered"] - shift
                )

        if all(col in df.columns for col in ["before_group_centered", "after_group_centered"]):
            df = df[(df["before_group_centered"] < 0) | (df["after_group_centered"] > 0)]
        elif df.columns[-1] == "before_group_centered":
            df = df[df["before_group_centered"] < 0]
        elif df.columns[-1] == "after_group_centered":
            df = df[df["after_group_centered"] > 0]

        df = df.reset_index()

        updated = source
        if not df.empty:
            updated = source[~source[event_id].isin(df[event_id])]

        collect_data_performance(
            scope="truncate_paths",
            event_name="metadata",
            called_params=self.to_dict()["values"],
            not_hash_values=["occurrence_before", "occurrence_after"],
            performance_data={
                "parent": {
                    "shape": source.shape,
                    "hash": hash_dataframe(source),
                },
                "child": {
                    "shape": updated.shape,
                    "hash": hash_dataframe(updated),
                },
            },
        )

        return updated



"""
Semantic caching for Natural Language Query system.

Uses Redis with embedding similarity matching for intelligent caching.
Caches both query results and LLM responses.
"""

from typing import Any, Dict, Optional, Tuple
import json
import hashlib
import os


class CacheManager:
    """
    Semantic cache using Redis and embedding similarity.

    Caches query results based on semantic similarity rather than exact matches.
    Uses sentence embeddings to find similar queries.
    """

    def __init__(
        self,
        redis_url: Optional[str] = None,
        ttl: int = 3600,
        similarity_threshold: float = 0.85,
        use_redis: bool = True,
    ):
        """
        Initialize cache manager.

        Args:
            redis_url: Redis connection URL (default: from env or localhost)
            ttl: Time-to-live for cache entries in seconds
            similarity_threshold: Minimum similarity score for cache hits (0-1)
            use_redis: If False, use in-memory cache (for testing)
        """
        self.ttl = ttl
        self.similarity_threshold = similarity_threshold
        self.use_redis = use_redis

        # In-memory fallback cache
        self._memory_cache: Dict[str, Any] = {}

        if use_redis:
            try:
                import redis
                from redis import Redis

                if redis_url is None:
                    redis_url = os.getenv("REDIS_URL", "redis://localhost:6379")

                self.redis_client: Optional[Redis] = redis.from_url(
                    redis_url, decode_responses=True
                )

                # Test connection
                self.redis_client.ping()
                self.redis_available = True
                print("Redis cache connected successfully")

            except (ImportError, Exception) as e:
                print(
                    f"Redis not available, using in-memory cache: {e}"
                )
                self.redis_client = None
                self.redis_available = False
        else:
            self.redis_client = None
            self.redis_available = False

    def _compute_query_hash(self, query: str) -> str:
        """
        Compute hash for query (for exact matching).

        Args:
            query: User query

        Returns:
            str: Hash string
        """
        return hashlib.sha256(query.lower().strip().encode()).hexdigest()

    def _get_embedding(self, text: str) -> list:
        """
        Get text embedding for semantic similarity.

        Args:
            text: Text to embed

        Returns:
            list: Embedding vector
        """
        try:
            from sentence_transformers import SentenceTransformer

            # Use a lightweight model
            model = SentenceTransformer("all-MiniLM-L6-v2")
            embedding = model.encode(text, convert_to_tensor=False)
            return embedding.tolist()

        except ImportError:
            # Fallback to hash-based matching if sentence-transformers not available
            return []

    def _cosine_similarity(self, vec1: list, vec2: list) -> float:
        """
        Calculate cosine similarity between two vectors.

        Args:
            vec1: First vector
            vec2: Second vector

        Returns:
            float: Similarity score (0-1)
        """
        if not vec1 or not vec2:
            return 0.0

        import numpy as np

        vec1 = np.array(vec1)
        vec2 = np.array(vec2)

        dot_product = np.dot(vec1, vec2)
        norm1 = np.linalg.norm(vec1)
        norm2 = np.linalg.norm(vec2)

        if norm1 == 0 or norm2 == 0:
            return 0.0

        return float(dot_product / (norm1 * norm2))

    def get(self, query: str) -> Optional[Dict[str, Any]]:
        """
        Get cached result for query.

        Uses semantic similarity to find similar queries.

        Args:
            query: User query

        Returns:
            Optional[dict]: Cached result or None
        """
        query_hash = self._compute_query_hash(query)

        # Try exact match first
        if self.redis_available and self.redis_client:
            try:
                cached = self.redis_client.get(f"nlq:exact:{query_hash}")
                if cached:
                    return json.loads(cached)
            except Exception as e:
                print(f"Redis get error: {e}")

        # Try in-memory cache
        if query_hash in self._memory_cache:
            return self._memory_cache[query_hash]

        # Try semantic similarity match
        if self.redis_available and self.redis_client:
            try:
                query_embedding = self._get_embedding(query)
                if query_embedding:
                    # Search for similar queries
                    cached_queries = self.redis_client.keys("nlq:semantic:*")

                    best_match = None
                    best_similarity = 0.0

                    for key in cached_queries:
                        try:
                            cached_data = json.loads(
                                self.redis_client.get(key)
                            )
                            cached_embedding = cached_data.get("embedding")

                            if cached_embedding:
                                similarity = self._cosine_similarity(
                                    query_embedding, cached_embedding
                                )

                                if (
                                    similarity > best_similarity
                                    and similarity >= self.similarity_threshold
                                ):
                                    best_similarity = similarity
                                    best_match = cached_data.get("result")

                        except Exception:
                            continue

                    if best_match:
                        print(
                            f"Cache hit (semantic similarity: {best_similarity:.2f})"
                        )
                        return best_match

            except Exception as e:
                print(f"Semantic search error: {e}")

        return None

    def set(self, query: str, result: Dict[str, Any]) -> None:
        """
        Cache query result.

        Args:
            query: User query
            result: Query result to cache
        """
        query_hash = self._compute_query_hash(query)

        # Store exact match
        if self.redis_available and self.redis_client:
            try:
                self.redis_client.setex(
                    f"nlq:exact:{query_hash}",
                    self.ttl,
                    json.dumps(result, default=str),
                )

                # Store with embedding for semantic search
                query_embedding = self._get_embedding(query)
                if query_embedding:
                    semantic_data = {
                        "query": query,
                        "embedding": query_embedding,
                        "result": result,
                    }
                    self.redis_client.setex(
                        f"nlq:semantic:{query_hash}",
                        self.ttl,
                        json.dumps(semantic_data, default=str),
                    )

            except Exception as e:
                print(f"Redis set error: {e}")

        # Also store in memory cache as fallback
        self._memory_cache[query_hash] = result

    def clear(self) -> None:
        """Clear all cached data."""
        if self.redis_available and self.redis_client:
            try:
                # Clear all NLQ cache keys
                keys = self.redis_client.keys("nlq:*")
                if keys:
                    self.redis_client.delete(*keys)
            except Exception as e:
                print(f"Redis clear error: {e}")

        # Clear memory cache
        self._memory_cache.clear()

    def get_stats(self) -> Dict[str, Any]:
        """
        Get cache statistics.

        Returns:
            dict: Cache stats
        """
        stats = {
            "redis_available": self.redis_available,
            "memory_cache_size": len(self._memory_cache),
            "ttl": self.ttl,
            "similarity_threshold": self.similarity_threshold,
        }

        if self.redis_available and self.redis_client:
            try:
                exact_keys = len(self.redis_client.keys("nlq:exact:*"))
                semantic_keys = len(self.redis_client.keys("nlq:semantic:*"))

                stats["redis_exact_cache_size"] = exact_keys
                stats["redis_semantic_cache_size"] = semantic_keys
                stats["redis_total_size"] = exact_keys + semantic_keys

            except Exception as e:
                stats["redis_error"] = str(e)

        return stats


def test_cache():
    """Test the cache manager."""
    cache = CacheManager(use_redis=False)

    # Test 1: Store and retrieve
    print("Test 1: Store and retrieve")
    query1 = "What is the conversion rate?"
    result1 = {"answer": "3.5%", "code": "..."}

    cache.set(query1, result1)
    retrieved = cache.get(query1)

    print(f"Stored: {result1}")
    print(f"Retrieved: {retrieved}")
    print(f"Match: {retrieved == result1}")
    print()

    # Test 2: Similar query (semantic matching)
    print("Test 2: Semantic matching")
    query2 = "What's the conversion percentage?"  # Similar to query1
    retrieved2 = cache.get(query2)

    print(f"Original query: {query1}")
    print(f"Similar query: {query2}")
    print(f"Retrieved: {retrieved2}")
    print()

    # Test 3: Stats
    print("Test 3: Cache stats")
    stats = cache.get_stats()
    print(json.dumps(stats, indent=2))


if __name__ == "__main__":
    test_cache()



"""
Natural Language Query Engine for Outhad_Edge.

Main orchestrator that combines all AI components to provide natural language
query capabilities for eventstream analysis.

 Implementation:
- Integrates LLMAgent 
- Uses execute_with_llm_retry 
- Uses CodeValidator integration 
- Supports visualization and insights
- Enhanced caching with retrieve_context
- Production-ready with comprehensive error handling
"""

from typing import Any, Dict, Optional, Literal
import os
import warnings
import logging

logger = logging.getLogger(__name__)

# Load environment variables from .env file
try:
    from dotenv import load_dotenv
    load_dotenv()
except ImportError:
    # python-dotenv not installed, will rely on system environment variables
    pass

# Disable ChromaDB telemetry to avoid compatibility errors
os.environ.setdefault('ANONYMIZED_TELEMETRY', 'False')


class NaturalLanguageQuery:
    """
    Natural Language Query engine for Eventstream analysis.

    Combines LLM code generation, vector search, semantic caching, and safe execution
    to enable natural language queries over eventstream data.

    Example:
        >>> stream = Eventstream(df)
        >>> nlq = stream.nlq(model="gpt-4")
        >>> result = nlq.ask("What is the conversion rate?")
        >>> print(result['answer'])
    """

    def __init__(
        self,
        eventstream,
        model: Literal["gpt-4", "gpt-3.5-turbo", "claude-3-sonnet", "claude-3-opus"] = "gpt-4",
        temperature: float = 0.1,
        max_retries: int = 3,
        use_cache: bool = True,
        verbose: bool = False,
        enable_monitoring: bool = True,
    ):
        """
        Initialize Natural Language Query engine.

        Args:
            eventstream: Outhad_Edge Eventstream object
            model: LLM model to use
            temperature: LLM temperature (0-1, lower = more deterministic)
            max_retries: Maximum retries for code generation/execution
            use_cache: Enable semantic caching
            verbose: Print debug information
            enable_monitoring: Enable query monitoring and logging 
        """
        self.eventstream = eventstream
        self.model = model
        self.temperature = temperature
        self.max_retries = max_retries
        self.use_cache = use_cache
        self.verbose = verbose
        self.enable_monitoring = enable_monitoring

        # Initialize components
        self._initialize_components()

    def _initialize_components(self) -> None:
        """Initialize all AI components ."""
        from .semantic_layer import SemanticLayer
        from .vector_store import VectorStore
        from .cache_manager import CacheManager

        # Semantic layer for schema understanding 
        self.semantic_layer = SemanticLayer(self.eventstream)
        if self.verbose:
            logger.info("Initialized SemanticLayer")

        # Vector store for RAG 
        try:
            self.vector_store = VectorStore()
            if self.vector_store.count() == 0:
                from .vector_store import initialize_vector_store
                self.vector_store = initialize_vector_store()
            if self.verbose:
                logger.info(f"Initialized VectorStore ({self.vector_store.count()} examples)")
        except Exception as e:
            warnings.warn(f"Vector store initialization failed: {e}")
            self.vector_store = None

        # Code executor with  enhancements
        from .code_executor import CodeExecutor
        self.code_executor = CodeExecutor(
            timeout_seconds=60,
            use_validator=True,  # : Use CodeValidator
            strict_mode=True
        )
        if self.verbose:
            logger.info("Initialized CodeExecutor (with CodeValidator)")

        # Cache manager
        if self.use_cache:
            try:
                self.cache_manager = CacheManager(use_redis=True)
            except Exception:
                warnings.warn("Redis cache not available, using in-memory cache")
                self.cache_manager = CacheManager(use_redis=False)
        else:
            self.cache_manager = None

        # Initialize LLM Agent 
        self._initialize_llm_agent()

        # Initialize monitoring 
        if self.enable_monitoring:
            try:
                from .monitoring import NLQMonitoring
                self.monitor = NLQMonitoring(enable_console=self.verbose)
                if self.verbose:
                    logger.info("Initialized NLQMonitoring ")
            except Exception as e:
                warnings.warn(f"Monitoring initialization failed: {e}")
                self.monitor = None
        else:
            self.monitor = None

    def _initialize_llm_agent(self) -> None:
        """Initialize LLM Agent ."""
        try:
            from .llm_agent import LLMAgent

            # Use LLMAgent instead of direct LLM 
            self.llm_agent = LLMAgent(
                model=self.model,
                temperature=self.temperature,
                verbose=self.verbose
            )

            if self.verbose:
                logger.info(f"Initialized LLMAgent (model={self.model}, temp={self.temperature})")

        except Exception as e:
            raise RuntimeError(
                f"Failed to initialize LLM Agent: {e}\n"
                "Make sure API keys are set in .env file:\n"
                "  - OPENAI_API_KEY for GPT models\n"
                "  - ANTHROPIC_API_KEY for Claude models"
            )

    def ask(self, question: str, return_code: bool = False, explain: bool = True) -> Dict[str, Any]:
        """
        Ask a natural language question about the eventstream.

        Args:
            question: Natural language question
            return_code: If True, include generated code in response
            explain: If True, include insights and explanation in response 

        Returns:
            dict: Query result with 'answer', 'result', 'code' (optional), 'cached', 'insights' (optional)
        """
        import time
        start_time = time.time()

        if self.verbose:
            print(f"Question: {question}")

        # Check cache first
        cached = False
        if self.cache_manager:
            cached_result = self.cache_manager.get(question)
            if cached_result:
                if self.verbose:
                    print("✓ Retrieved from cache")
                cached_result["cached"] = True
                cached = True

                # Log cached query 
                if self.monitor:
                    latency = time.time() - start_time
                    self.monitor.log_query(
                        query=question,
                        success=True,
                        latency=latency,
                        model=self.model,
                        tokens_used=0,
                        cached=True,
                    )

                return cached_result

        # Get semantic context
        context = self.semantic_layer.get_full_context()

        # Retrieve similar examples using enhanced RAG 
        examples = ""
        if self.vector_store:
            try:
                # Use retrieve_context for better RAG ( enhancement)
                examples = self.vector_store.retrieve_context(question, n_results=2)
                if self.verbose:
                    print(f"✓ Retrieved {len(examples.split('Example'))-1 if examples else 0} similar examples")
            except Exception as e:
                if self.verbose:
                    print(f"Vector search failed: {e}")
                # Fallback to basic search
                try:
                    similar = self.vector_store.search(question, n_results=2)
                    examples = "\n\n".join(
                        [
                            f"Q: {ex['question']}\nCode:\n```python\n{ex['code']}\n```"
                            for ex in similar
                        ]
                    )
                except Exception:
                    examples = ""

        # Generate code
        code = self._generate_code(question, context, examples)

        if self.verbose:
            print(f"Generated code:\n{code}\n")

        # Execute code with retry
        exec_result = self._execute_with_retry(code, question)

        if not exec_result["success"]:
            # Log failed query 
            if self.monitor:
                latency = time.time() - start_time
                self.monitor.log_query(
                    query=question,
                    success=False,
                    latency=latency,
                    model=self.model,
                    tokens_used=0,
                    cached=False,
                    error=exec_result["error"],
                )

            return {
                "success": False,
                "error": exec_result["error"],
                "answer": f"Failed to answer: {exec_result['error']}",
                "question": question,
                "code": code if return_code else None,
                "cached": False,
            }

        # Format answer
        answer = self._format_answer(question, exec_result["result"], code)

        result = {
            "success": True,
            "answer": answer,
            "result": exec_result["result"],
            "question": question,
            "cached": False,
        }

        if return_code:
            result["code"] = code

        # Add insights and visualization support 
        if explain:
            insights = self._generate_insights(question, exec_result["result"], code)
            result["insights"] = insights
            result["visualization"] = self._suggest_visualization(exec_result["result"])

        # Cache result
        if self.cache_manager:
            self.cache_manager.set(question, result)

        # Log successful query 
        if self.monitor:
            latency = time.time() - start_time
            # Estimate tokens (rough approximation: ~4 chars per token)
            tokens_estimate = (len(question) + len(str(code)) + len(str(answer))) // 4
            self.monitor.log_query(
                query=question,
                success=True,
                latency=latency,
                model=self.model,
                tokens_used=tokens_estimate,
                cached=False,
            )

        return result

    def _generate_code(
        self, question: str, context: str, examples: str
    ) -> str:
        """Generate Python code to answer the question ( uses LLMAgent)."""
        try:
            # Use LLMAgent instead of direct LLM 
            code = self.llm_agent.generate_code(
                query=question,
                schema=context,
                examples=examples,
                glossary=""  # Can be enhanced later with domain glossary
            )
            return code

        except Exception as e:
            raise RuntimeError(f"Code generation failed: {e}")

    def _execute_with_retry(
        self, code: str, question: str
    ) -> Dict[str, Any]:
        """Execute code with retry and LLM-based error correction ( uses  method)."""
        # Use execute_with_llm_retry from  (integrates CodeValidator and LLMAgent)
        success, result, error = self.code_executor.execute_with_llm_retry(
            query=question,
            initial_code=code,
            stream=self.eventstream,
            llm_agent=self.llm_agent,
            max_retries=self.max_retries,
            verbose=self.verbose
        )

        if success:
            return result
        else:
            return {"success": False, "error": error}

    def _fix_code(self, code: str, error: str, question: str) -> str:
        """
        Use LLM to fix code error.

        Note: This method is deprecated in . Error correction is now handled
        by execute_with_llm_retry() which uses LLMAgent.correct_error().
        Kept for backward compatibility.
        """
        try:
            # Use LLMAgent for error correction 
            fixed_code = self.llm_agent.correct_error(
                query=question,
                code=code,
                error=error
            )
            return fixed_code

        except Exception as e:
            if self.verbose:
                print(f"Code correction failed: {e}")
            return code  # Return original code if correction fails

    def _format_answer(
        self, question: str, raw_result: Any, code: str
    ) -> str:
        """Format raw result into natural language answer."""
        # Simple formatting for now
        if raw_result is None:
            return "No result generated"

        if isinstance(raw_result, (int, float)):
            return f"{raw_result}"

        if isinstance(raw_result, str):
            return raw_result

        if hasattr(raw_result, "to_string"):
            return raw_result.to_string()

        return str(raw_result)

    def _generate_insights(self, question: str, result: Any, code: str) -> str:
        """
        Generate insights about the result .

        Args:
            question: User's query
            result: Execution result
            code: Generated code

        Returns:
            str: Human-readable insights
        """
        insights = []

        # Basic result type insight
        if isinstance(result, (int, float)):
            insights.append(f"The result is a numeric value: {result}")
        elif hasattr(result, 'shape'):
            # DataFrame or numpy array
            insights.append(f"Result shape: {result.shape}")
            if hasattr(result, 'columns'):
                insights.append(f"Columns: {', '.join(str(c) for c in result.columns)}")
        elif isinstance(result, (list, tuple)):
            insights.append(f"Result is a sequence with {len(result)} items")

        # Code complexity insight
        lines = code.strip().split('\n')
        non_empty_lines = [l for l in lines if l.strip() and not l.strip().startswith('#')]
        insights.append(f"Generated code: {len(non_empty_lines)} lines")

        return "\n".join(insights) if insights else "No additional insights available"

    def _suggest_visualization(self, result: Any) -> Optional[str]:
        """
        Suggest appropriate visualization type based on result .

        Args:
            result: Execution result

        Returns:
            str: Suggested visualization type or None
        """
        if isinstance(result, (int, float)):
            return "metric"  # Single metric display

        if hasattr(result, 'shape'):
            # DataFrame-like result
            if hasattr(result, 'columns'):
                num_cols = len(result.columns)
                num_rows = len(result)

                if num_rows <= 20:
                    return "table"  # Small enough to show as table
                elif num_cols == 2:
                    return "bar_chart"  # Two columns: category and value
                elif num_cols >= 3:
                    return "line_chart"  # Time series or multi-column data

        if isinstance(result, (list, tuple)):
            if len(result) <= 10:
                return "list"  # Simple list display
            else:
                return "table"  # Convert to table

        return None  # No specific visualization suggested

    def get_stats(self) -> Dict[str, Any]:
        """Get statistics about the NLQ engine."""
        stats = {
            "model": self.model,
            "temperature": self.temperature,
            "cache_enabled": self.use_cache,
            "vector_store_size": (
                self.vector_store.count() if self.vector_store else 0
            ),
        }

        if self.cache_manager:
            stats["cache_stats"] = self.cache_manager.get_stats()

        return stats

    def clear_cache(self) -> None:
        """Clear the semantic cache."""
        if self.cache_manager:
            self.cache_manager.clear()
            if self.verbose:
                print("Cache cleared")



from __future__ import annotations

import json
from pathlib import Path
from typing import Any, Literal, Optional

from ipykernel.comm.comm import Comm

from outhad_edge.backend import JupyterServer
from outhad_edge.exceptions.server import ServerErrorWithResponse
from outhad_edge.utils.singleton import Singleton

EnvId = Literal["classic", "colab"]


class ServerManager:
    __metaclass__ = Singleton
    kernel_id: Optional[str]
    _servers: dict[str, JupyterServer]

    def __init__(self) -> None:
        self._servers: dict[str, JupyterServer] = {}
        self._main_listener_created = False
        self.kernel_id = self._get_kernel_id()

    def _get_kernel_id(self) -> str | None:
        try:
            import ipykernel

            connection_file = Path(ipykernel.get_connection_file()).stem
            return connection_file.split("-", 1)[1]
        except:
            return None

    def _find_server(self, server_id: str) -> JupyterServer | None:
        return self._servers.get(server_id, None)

    def _on_colab_func_called(self, server_id: str, method: str, request_id: str, payload: dict) -> str:
        target_server: JupyterServer | None = self._find_server(server_id)
        if target_server is None:
            err = "ServerNotFound"
            return json.dumps(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": str(err),
                }
            )
        try:
            result = target_server.dispatch_method(method=method, payload=payload)
            return json.dumps(
                {
                    "success": True,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": result,
                }
            )
        except ServerErrorWithResponse as err:
            return json.dumps(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": err.dict(),
                }
            )
        except Exception as err:
            wrapped_exc = ServerErrorWithResponse(message=str(err), type="unexpected_error")

            return json.dumps(
                {
                    "success": False,
                    "server_id": server_id,
                    "request_id": request_id,
                    "method": method,
                    "result": wrapped_exc.dict(),
                }
            )

    def _on_comm_message(self, comm: Comm, open_msg: Any) -> None:
        @comm.on_msg  # type: ignore
        def _recv(msg: dict[str, dict]) -> None:
            data: dict[str, Any] = msg["content"]["data"]
            server_id = data["server_id"]
            request_id = data["request_id"]
            request_type = data.get("request_type", None)

            target_server = self._find_server(server_id)
            if target_server is None:
                raise Exception("server not found!")

            if request_type == "handshake":
                comm.send(
                    {"success": True, "server_id": server_id, "request_id": request_id, "request_type": "handshake"}
                )
                return

            method = data["method"]
            payload = data.get("payload", {})

            try:
                result = target_server.dispatch_method(method=method, payload=payload)
                comm.send(
                    {
                        "success": True,
                        "server_id": server_id,
                        "request_id": request_id,
                        "method": method,
                        "result": result,
                    }
                )
            except ServerErrorWithResponse as err:
                comm.send(
                    {
                        "success": False,
                        "server_id": server_id,
                        "request_id": request_id,
                        "method": method,
                        "result": err.dict(),
                    }
                )
            except Exception as err:
                wrapped_exc = ServerErrorWithResponse(message=str(err), type="unexpected_error")

                comm.send(
                    {
                        "success": False,
                        "server_id": server_id,
                        "request_id": request_id,
                        "method": method,
                        "result": wrapped_exc.dict(),
                    }
                )

    def _create_main_listener(self) -> None:
        env = self.check_env()

        if env == "colab":
            import google.colab.output  # type: ignore

            google.colab.output.register_callback(
                "JupyterServerMainCallback",
                lambda server_id, method, request_id, payload: self._on_colab_func_called(
                    server_id, method, request_id, payload
                ),
            )
        if env == "classic":
            from IPython.core.getipython import get_ipython

            if ipython := get_ipython():
                ipython.kernel.comm_manager.register_target(
                    "JupyterServerMainCallback", lambda comm, open_msg: self._on_comm_message(comm, open_msg)
                )
        self._main_listener_created = True

    def check_env(self) -> EnvId:
        try:
            import google.colab  # type: ignore # noqa: F401

            return "colab"
        except ImportError:
            return "classic"

    def create_server(self, pk: Optional[str] = None) -> JupyterServer:
        server = JupyterServer(pk=pk, kernel_id=self.kernel_id)
        self._create_main_listener()
        self._servers[server.pk] = server
        return server



from __future__ import annotations

from abc import abstractmethod
from dataclasses import field
from typing import Any, List, Optional, Protocol, TypedDict

import pandas as pd

IndexOrder = List[Optional[str]]


class EventstreamType(Protocol):
    schema: EventstreamSchemaType
    index_order: IndexOrder
    __raw_data_schema: RawDataSchemaType
    __events: pd.DataFrame | pd.Series[Any]

    @abstractmethod
    def copy(self) -> EventstreamType:
        ...

    @property
    @abstractmethod
    def _eventstream_index(self) -> int:
        ...

    @abstractmethod
    def append_eventstream(self, eventstream: EventstreamType) -> None:
        ...

    @property
    @abstractmethod
    def _hash(self) -> str:
        ...

    @abstractmethod
    def to_dataframe(self, copy: bool = False) -> pd.DataFrame:
        ...

    @abstractmethod
    def add_custom_col(self, name: str, data: pd.Series[Any] | None) -> None:
        ...


class EventstreamSchemaType(Protocol):
    custom_cols: List[str] = field(default_factory=list)
    user_id: str = "user_id"
    event_timestamp: str = "event_timestamp"
    event_name: str = "event_name"
    event_index: str = "event_index"
    event_type: str = "event_type"
    event_id: str = "event_id"

    @abstractmethod
    def copy(self) -> EventstreamSchemaType:
        ...

    @abstractmethod
    def is_equal(self, schema: EventstreamSchemaType) -> bool:
        ...

    @abstractmethod
    def get_default_cols(self) -> List[str]:
        ...

    @abstractmethod
    def get_cols(self) -> list[str]:
        ...

    @abstractmethod
    def to_raw_data_schema(self, event_id: bool = False, event_index: bool = False) -> RawDataSchemaType:
        ...


class RawDataCustomColSchema(TypedDict):
    raw_data_col: str
    custom_col: str


class RawDataSchemaType(Protocol):
    event_name: str = "event"
    event_timestamp: str = "timestamp"
    user_id: str = "user_id"
    event_index: Optional[str] = None
    event_type: Optional[str] = None
    event_id: Optional[str] = None
    custom_cols: List[RawDataCustomColSchema] = field(default_factory=list)

    @abstractmethod
    def get_default_cols(self) -> List[str]:
        ...

    @abstractmethod
    def copy(self) -> RawDataSchemaType:
        ...



"""
LLM Agent for code generation using LangChain.

Provides a clean abstraction over different LLM providers (OpenAI, Anthropic)
with built-in retry logic, error correction, and production-grade features.
"""

from typing import Optional, Dict, Any, Literal
from functools import wraps
import time
import logging
import os

logger = logging.getLogger(__name__)


def retry_on_llm_error(max_retries: int = 3, delay: float = 1.0):
    """
    Decorator to retry LLM calls on failure with exponential backoff.

    Args:
        max_retries: Maximum number of retry attempts
        delay: Initial delay in seconds (doubles with each retry)
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            last_exception = None
            for attempt in range(max_retries):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    last_exception = e
                    if attempt < max_retries - 1:
                        wait_time = delay * (2 ** attempt)
                        logger.warning(
                            f"LLM call failed (attempt {attempt + 1}/{max_retries}): {e}. "
                            f"Retrying in {wait_time}s..."
                        )
                        time.sleep(wait_time)
                    else:
                        logger.error(f"LLM call failed after {max_retries} attempts: {e}")
            raise last_exception
        return wrapper
    return decorator


class LLMAgent:
    """
    LLM agent for generating code from natural language using LangChain.

    Supports multiple LLM providers with consistent interface:
    - OpenAI (GPT-4, GPT-3.5-turbo)
    - Anthropic (Claude-3-Sonnet, Claude-3-Opus)

    Features:
    - Automatic retry with exponential backoff
    - Error correction capabilities
    - Token usage tracking
    - Verbose mode for debugging

    Example:
        >>> agent = LLMAgent(model="gpt-4", temperature=0.1)
        >>> code = agent.generate_code(
        ...     query="What is the conversion rate?",
        ...     schema="user_id, event, timestamp",
        ...     examples="...",
        ...     glossary="conversion: users who purchased"
        ... )
    """

    SUPPORTED_PROVIDERS = {
        "gpt-4": "openai",
        "gpt-3.5-turbo": "openai",
        "gpt-4-turbo": "openai",
        "claude-3-sonnet": "anthropic",
        "claude-3-opus": "anthropic",
        "claude-3-haiku": "anthropic",
    }

    def __init__(
        self,
        model: str = "gpt-4",
        temperature: float = 0.1,
        provider: Optional[Literal["openai", "anthropic"]] = None,
        max_tokens: Optional[int] = None,
        verbose: bool = False,
    ):
        """
        Initialize LLM agent.

        Args:
            model: Model name (gpt-4, gpt-3.5-turbo, claude-3-sonnet, etc.)
            temperature: Sampling temperature (0-1). Lower = more deterministic
            provider: LLM provider (auto-detected if None)
            max_tokens: Maximum tokens in response (provider default if None)
            verbose: Enable verbose logging

        Raises:
            ValueError: If model is unsupported or API key is missing
            ImportError: If required LangChain package is not installed
        """
        self.model = model
        self.temperature = temperature
        self.verbose = verbose
        self.max_tokens = max_tokens

        # Auto-detect provider if not specified
        if provider is None:
            provider = self._detect_provider(model)
        self.provider = provider

        # Validate temperature
        if not 0 <= temperature <= 1:
            raise ValueError(f"Temperature must be between 0 and 1, got {temperature}")

        # Initialize LLM
        self.llm = self._initialize_llm()

        # Track usage
        self.total_tokens_used = 0
        self.calls_made = 0

        if verbose:
            logger.setLevel(logging.DEBUG)
            logger.debug(f"Initialized LLMAgent with {model} (provider: {provider})")

    def _detect_provider(self, model: str) -> str:
        """Auto-detect provider from model name."""
        if model in self.SUPPORTED_PROVIDERS:
            return self.SUPPORTED_PROVIDERS[model]

        # Fallback: guess by model name prefix
        if model.startswith("gpt"):
            return "openai"
        elif model.startswith("claude"):
            return "anthropic"
        else:
            raise ValueError(
                f"Unknown model: {model}. Supported models: {list(self.SUPPORTED_PROVIDERS.keys())}"
            )

    def _initialize_llm(self):
        """Initialize LLM client based on provider."""
        if self.provider == "openai":
            return self._initialize_openai()
        elif self.provider == "anthropic":
            return self._initialize_anthropic()
        else:
            raise ValueError(f"Unknown provider: {self.provider}")

    def _initialize_openai(self):
        """Initialize OpenAI LLM client."""
        try:
            from langchain_openai import ChatOpenAI
        except ImportError:
            raise ImportError(
                "langchain-openai required for OpenAI models. "
                "Install with: pip install langchain-openai"
            )

        api_key = os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError(
                "OPENAI_API_KEY environment variable not set. "
                "Add it to your .env file or set it in your environment."
            )

        kwargs = {
            "model": self.model,
            "temperature": self.temperature,
            "api_key": api_key,
        }

        if self.max_tokens:
            kwargs["max_tokens"] = self.max_tokens

        return ChatOpenAI(**kwargs)

    def _initialize_anthropic(self):
        """Initialize Anthropic LLM client."""
        try:
            from langchain_anthropic import ChatAnthropic
        except ImportError:
            raise ImportError(
                "langchain-anthropic required for Anthropic models. "
                "Install with: pip install langchain-anthropic"
            )

        api_key = os.getenv("ANTHROPIC_API_KEY")
        if not api_key:
            raise ValueError(
                "ANTHROPIC_API_KEY environment variable not set. "
                "Add it to your .env file or set it in your environment."
            )

        # Map friendly names to actual model IDs
        model_map = {
            "claude-3-sonnet": "claude-3-sonnet-20240229",
            "claude-3-opus": "claude-3-opus-20240229",
            "claude-3-haiku": "claude-3-haiku-20240307",
        }

        model_id = model_map.get(self.model, self.model)

        kwargs = {
            "model": model_id,
            "temperature": self.temperature,
            "api_key": api_key,
        }

        if self.max_tokens:
            kwargs["max_tokens"] = self.max_tokens

        return ChatAnthropic(**kwargs)

    @retry_on_llm_error(max_retries=3, delay=1.0)
    def generate_code(
        self,
        query: str,
        schema: str,
        examples: str,
        glossary: str,
        system_prompt: Optional[str] = None,
        custom_instructions: Optional[str] = None,
    ) -> str:
        """
        Generate Python code from natural language query.

        Args:
            query: User's natural language question
            schema: Schema description (columns, data types, stats)
            examples: Similar example queries with code
            glossary: Business term definitions
            system_prompt: Custom system prompt (uses default if None)
            custom_instructions: Additional instructions to append

        Returns:
            Generated Python code as string

        Raises:
            ValueError: If inputs are invalid
            RuntimeError: If LLM call fails after retries
        """
        # Validate inputs
        if not query or not query.strip():
            raise ValueError("query cannot be empty")

        if self.verbose:
            logger.debug(f"Generating code for query: {query[:100]}...")

        # Use default system prompt if not provided
        if system_prompt is None:
            from .prompt_templates import PromptTemplates
            system_prompt = PromptTemplates.get_system_prompt()

        # Create user prompt
        from .prompt_templates import PromptTemplates
        user_prompt = PromptTemplates.get_user_prompt(
            query=query,
            schema=schema,
            examples=examples,
            glossary=glossary
        )

        # Add custom instructions if provided
        if custom_instructions:
            user_prompt += f"\n\n**Additional Instructions:**\n{custom_instructions}"

        # Create chat messages
        from langchain.prompts import ChatPromptTemplate

        prompt = ChatPromptTemplate.from_messages([
            ("system", system_prompt),
            ("user", user_prompt)
        ])

        # Generate code
        chain = prompt | self.llm
        response = chain.invoke({})

        # Track usage
        self.calls_made += 1
        if hasattr(response, 'response_metadata'):
            usage = response.response_metadata.get('token_usage', {})
            tokens = usage.get('total_tokens', 0)
            self.total_tokens_used += tokens
            if self.verbose:
                logger.debug(f"Tokens used: {tokens} (total: {self.total_tokens_used})")

        # Extract code from response
        code = self._extract_code(response.content)

        if self.verbose:
            logger.debug(f"Generated code ({len(code)} chars)")

        return code

    @retry_on_llm_error(max_retries=2, delay=0.5)
    def correct_error(
        self,
        query: str,
        code: str,
        error: str,
        attempt: int = 1
    ) -> str:
        """
        Correct code that failed execution.

        Args:
            query: Original natural language query
            code: Failed Python code
            error: Error message from execution
            attempt: Current correction attempt number

        Returns:
            Corrected Python code

        Raises:
            RuntimeError: If correction fails after retries
        """
        if self.verbose:
            logger.debug(f"Correcting error (attempt {attempt}): {error[:100]}...")

        # Create error correction prompt
        from .prompt_templates import PromptTemplates
        prompt_text = PromptTemplates.get_error_correction_prompt(
            query=query,
            code=code,
            error=error
        )

        from langchain.prompts import ChatPromptTemplate

        prompt = ChatPromptTemplate.from_messages([
            ("system", "You are a Python debugging expert. Fix the code to handle the error."),
            ("user", prompt_text)
        ])

        # Get correction
        chain = prompt | self.llm
        response = chain.invoke({})

        # Track usage
        self.calls_made += 1

        # Extract corrected code
        corrected_code = self._extract_code(response.content)

        if self.verbose:
            logger.debug(f"Generated corrected code ({len(corrected_code)} chars)")

        return corrected_code

    def _extract_code(self, response: str) -> str:
        """
        Extract Python code from LLM response.

        Handles markdown code blocks and other formatting.
        """
        code = response.strip()

        # Remove markdown code blocks
        if code.startswith("```python"):
            code = code[len("```python"):].strip()
        elif code.startswith("```"):
            code = code[3:].strip()

        if code.endswith("```"):
            code = code[:-3].strip()

        return code

    def get_usage_stats(self) -> Dict[str, Any]:
        """
        Get usage statistics.

        Returns:
            dict: Statistics including tokens used, calls made, etc.
        """
        return {
            "model": self.model,
            "provider": self.provider,
            "calls_made": self.calls_made,
            "total_tokens_used": self.total_tokens_used,
            "avg_tokens_per_call": (
                self.total_tokens_used / self.calls_made
                if self.calls_made > 0
                else 0
            ),
        }

    def reset_stats(self) -> None:
        """Reset usage statistics."""
        self.total_tokens_used = 0
        self.calls_made = 0
        if self.verbose:
            logger.debug("Usage statistics reset")

    def __repr__(self) -> str:
        return (
            f"LLMAgent(model='{self.model}', provider='{self.provider}', "
            f"temperature={self.temperature}, calls={self.calls_made})"
        )



from __future__ import annotations

import warnings
from typing import Literal

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from outhad_edge.backend.tracker import (
    collect_data_performance,
    time_performance,
    track,
)
from outhad_edge.eventstream.types import EventstreamType
from outhad_edge.tooling.constants import BINS_ESTIMATORS


class EventTimestampHist:
    """
    Plot the distribution of events over time. Can be useful for detecting
    time-based anomalies, and visualising general timespan of the eventstream.

    Parameters
    ----------
    eventstream : EventstreamType

    See Also
    --------
    .TimedeltaHist : Plot the distribution of the time deltas between two events.
    .UserLifetimeHist : Plot the distribution of user lifetimes.
    .Eventstream.describe : Show general eventstream statistics.
    .Eventstream.describe_events : Show general eventstream events statistics.

    Notes
    -----
    See :ref:`Eventstream user guide<eventstream_events_timestamp>` for the details.
    """

    __eventstream: EventstreamType
    raw_events_only: bool
    event_list: list[str] | None
    lower_cutoff_quantile: float | None
    upper_cutoff_quantile: float | None
    bins: int | Literal[BINS_ESTIMATORS]
    bins_to_show: np.ndarray
    values_to_plot: np.ndarray

    @time_performance(
        scope="event_timestamp_hist",
        event_name="init",
    )
    def __init__(self, eventstream: EventstreamType) -> None:
        self.__eventstream = eventstream
        self.user_col = self.__eventstream.schema.user_id
        self.event_col = self.__eventstream.schema.event_name
        self.time_col = self.__eventstream.schema.event_timestamp

        self.bins_to_show = np.array([])
        self.values_to_plot = np.array([])

    def _remove_cutoff_values(self, series: pd.Series) -> pd.Series:
        idx = [True] * len(series)
        if self.upper_cutoff_quantile is not None:
            idx &= series <= series.quantile(self.upper_cutoff_quantile)
        if self.lower_cutoff_quantile is not None:
            idx &= series >= series.quantile(self.lower_cutoff_quantile)
        return series[idx]

    def __validate_input(
        self,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
    ) -> tuple[float | None, float | None]:
        if lower_cutoff_quantile is not None:
            if not 0 < lower_cutoff_quantile < 1:
                raise ValueError("lower_cutoff_quantile should be a fraction between 0 and 1.")

        if upper_cutoff_quantile is not None:
            if not 0 < upper_cutoff_quantile < 1:
                raise ValueError("upper_cutoff_quantile should be a fraction between 0 and 1.")

        if lower_cutoff_quantile is not None and upper_cutoff_quantile is not None:
            if lower_cutoff_quantile > upper_cutoff_quantile:
                warnings.warn("lower_cutoff_quantile exceeds upper_cutoff_quantile; no data passed to the histogram")

        return upper_cutoff_quantile, lower_cutoff_quantile

    @time_performance(
        scope="event_timestamp_hist",
        event_name="fit",
    )
    def fit(
        self,
        raw_events_only: bool = False,
        event_list: list[str] | None = None,
        lower_cutoff_quantile: float | None = None,
        upper_cutoff_quantile: float | None = None,
        bins: int | Literal[BINS_ESTIMATORS] = 20,
    ) -> None:
        """
        Calculate values for the histplot.

        Parameters
        ----------
        raw_events_only : bool, default False
            If ``True`` - statistics will only be shown for raw events.
            If ``False`` - statistics will be shown for all events presented in your data.
        event_list : list of str, optional
            Specify events to be displayed.
        lower_cutoff_quantile : float, optional
            Specify time distance quantile as the lower boundary. The values below the boundary are truncated.
        upper_cutoff_quantile : float, optional
            Specify time distance quantile as the upper boundary. The values above the boundary are truncated.
        bins : int or str, default 20
            Generic bin parameter that can be the name of a reference rule or
            the number of bins. Passed to :numpy_bins_link:`numpy.histogram_bin_edges<>`.

        Returns
        -------
        None

        """
        called_params = {
            "raw_events_only": raw_events_only,
            "event_list": event_list,
            "lower_cutoff_quantile": lower_cutoff_quantile,
            "upper_cutoff_quantile": upper_cutoff_quantile,
            "bins": bins,
        }

        self.upper_cutoff_quantile, self.lower_cutoff_quantile = self.__validate_input(
            lower_cutoff_quantile,
            upper_cutoff_quantile,
        )

        self.event_list = event_list
        self.raw_events_only = raw_events_only
        self.bins = bins

        data = self.__eventstream.to_dataframe(copy=True)

        if self.raw_events_only:
            data = data[data["event_type"].isin(["raw"])]
        if self.event_list:
            data = data[data[self.event_col].isin(self.event_list)]

        values_to_plot = data[self.time_col]
        if self._remove_cutoff_values:  # type: ignore
            values_to_plot = self._remove_cutoff_values(values_to_plot).to_numpy()

        bins_to_show = np.histogram_bin_edges(pd.to_numeric(values_to_plot), bins=self.bins)
        bins_to_show = pd.to_datetime(bins_to_show).round("s")
        if len(values_to_plot) == 0:
            bins_to_show = np.array([])

        self.bins_to_show = bins_to_show  # type: ignore
        self.values_to_plot = values_to_plot  # type: ignore

        collect_data_performance(
            scope="event_timestamp_hist",
            event_name="metadata",
            called_params=called_params,
            performance_data={},
            eventstream_index=self.__eventstream._eventstream_index,
        )

    @property
    @time_performance(  # type: ignore
        scope="event_timestamp_hist",
        event_name="values",
    )
    def values(self) -> tuple[np.ndarray, np.ndarray]:
        """

        Returns
        -------
        tuple(np.ndarray, np.ndarray)

            1. The first array contains the values for histogram.
            2. The first array contains the bin edges.

        """
        return self.values_to_plot, self.bins_to_show

    @time_performance(
        scope="event_timestamp_hist",
        event_name="plot",
    )
    def plot(self, width: float = 6.0, height: float = 4.5) -> matplotlib.axes.Axesne:
        """
        Create a sns.histplot based on the calculated values.

        Parameters
        ----------
        width : float, default 6.0
            Width in inches.
        height : float, default 4.5
            Height in inches.

        Returns
        -------
        :matplotlib_axes:`matplotlib.axes.Axes<>`
            The matplotlib axes containing the plot.

        """

        figsize = (width, height)
        plt.figure(figsize=figsize)

        hist = sns.histplot(self.values_to_plot, bins=self.bins)
        hist.set_title("Event timestamp histogram")

        collect_data_performance(
            scope="event_timestamp_hist",
            event_name="metadata",
            called_params={"width": width, "height": height},
            performance_data={},
            eventstream_index=self.__eventstream._eventstream_index,
        )

        return hist



# * Copyright (C) 2020 Mohammad Tanzil Idrisi, Outhad_Edge Team
# * This Source Code Form is subject to the terms of the Outhad_Edge Software Non-Exclusive License (License)
# * By using, sharing or editing this code you agree with the License terms and conditions.
# * You can obtain License text at https://github.com/Outhad-Lab/outhad_edge/blob/master/LICENSE.md
from .config import RETE_CONFIG
from .pandas import get_merged_col, shuffle_df  # type: ignore

__all__ = [
    "get_merged_col",
    "shuffle_df",
    "RETE_CONFIG",
]



"""
Monitoring and logging for Natural Language Query system.

Tracks query execution, performance metrics, and usage statistics.
Provides production-ready observability for the NLQ engine.
"""

import logging
import time
import json
from typing import Dict, Any, Optional, List
from pathlib import Path
from datetime import datetime
from collections import defaultdict


class NLQMonitoring:
    """
    Monitor NLQ system performance and usage.

    Tracks queries, latency, success rates, token usage, and cache performance.
    Provides metrics for production monitoring and optimization.
    """

    def __init__(
        self,
        log_file: Optional[str] = None,
        log_level: int = logging.INFO,
        enable_console: bool = False,
    ):
        """
        Initialize NLQ monitoring.

        Args:
            log_file: Path to log file (default: ./logs/nlq_queries.log)
            log_level: Logging level (default: INFO)
            enable_console: Also log to console (default: False)
        """
        self.logger = logging.getLogger("outhad_edge.nlq")
        self.logger.setLevel(log_level)

        # Remove existing handlers to avoid duplicates
        self.logger.handlers = []

        # Create logs directory if needed
        if log_file is None:
            log_dir = Path("./logs")
            log_dir.mkdir(exist_ok=True)
            log_file = str(log_dir / "nlq_queries.log")

        # File handler
        file_handler = logging.FileHandler(log_file)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        file_handler.setFormatter(formatter)
        self.logger.addHandler(file_handler)

        # Console handler (optional)
        if enable_console:
            console_handler = logging.StreamHandler()
            console_handler.setFormatter(formatter)
            self.logger.addHandler(console_handler)

        # In-memory metrics for aggregation
        self._metrics: Dict[str, Any] = {
            "total_queries": 0,
            "successful_queries": 0,
            "failed_queries": 0,
            "cache_hits": 0,
            "cache_misses": 0,
            "total_latency": 0.0,
            "total_tokens": 0,
            "queries_by_model": defaultdict(int),
            "errors_by_type": defaultdict(int),
        }

        self._query_history: List[Dict[str, Any]] = []
        self._max_history = 1000  # Keep last 1000 queries in memory

        self.logger.info("NLQ Monitoring initialized")

    def log_query(
        self,
        query: str,
        success: bool,
        latency: float,
        model: str = "unknown",
        tokens_used: int = 0,
        cached: bool = False,
        error: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """
        Log query execution.

        Args:
            query: User query text
            success: Whether query succeeded
            latency: Execution time in seconds
            model: LLM model used
            tokens_used: Number of tokens consumed
            cached: Whether result was from cache
            error: Error message if failed
            metadata: Additional metadata
        """
        log_data = {
            "timestamp": datetime.now().isoformat(),
            "query": query[:200],  # Truncate long queries
            "success": success,
            "latency_ms": round(latency * 1000, 2),
            "model": model,
            "tokens": tokens_used,
            "cached": cached,
        }

        if error:
            log_data["error"] = error

        if metadata:
            log_data.update(metadata)

        # Log as JSON for easy parsing
        log_message = json.dumps(log_data)

        if success:
            self.logger.info(log_message)
        else:
            self.logger.error(log_message)

        # Update metrics
        self._update_metrics(
            success=success,
            latency=latency,
            model=model,
            tokens=tokens_used,
            cached=cached,
            error=error,
        )

        # Store in history (limited size)
        self._query_history.append(log_data)
        if len(self._query_history) > self._max_history:
            self._query_history.pop(0)

    def _update_metrics(
        self,
        success: bool,
        latency: float,
        model: str,
        tokens: int,
        cached: bool,
        error: Optional[str],
    ) -> None:
        """
        Update in-memory metrics.

        Args:
            success: Whether query succeeded
            latency: Execution time in seconds
            model: LLM model used
            tokens: Token count
            cached: Whether cached
            error: Error message
        """
        self._metrics["total_queries"] += 1

        if success:
            self._metrics["successful_queries"] += 1
        else:
            self._metrics["failed_queries"] += 1

        if cached:
            self._metrics["cache_hits"] += 1
        else:
            self._metrics["cache_misses"] += 1

        self._metrics["total_latency"] += latency
        self._metrics["total_tokens"] += tokens
        self._metrics["queries_by_model"][model] += 1

        if error:
            # Extract error type
            error_type = error.split(":")[0] if ":" in error else "Unknown"
            self._metrics["errors_by_type"][error_type] += 1

    def get_metrics(self) -> Dict[str, Any]:
        """
        Get aggregated usage metrics.

        Returns:
            dict: Performance and usage statistics
        """
        total_queries = self._metrics["total_queries"]

        if total_queries == 0:
            return {
                "total_queries": 0,
                "success_rate": 0.0,
                "avg_latency_ms": 0.0,
                "cache_hit_rate": 0.0,
                "total_tokens": 0,
            }

        metrics = {
            "total_queries": total_queries,
            "successful_queries": self._metrics["successful_queries"],
            "failed_queries": self._metrics["failed_queries"],
            "success_rate": round(
                self._metrics["successful_queries"] / total_queries * 100, 2
            ),
            "cache_hits": self._metrics["cache_hits"],
            "cache_misses": self._metrics["cache_misses"],
            "cache_hit_rate": round(
                self._metrics["cache_hits"] / total_queries * 100, 2
            ),
            "avg_latency_ms": round(
                (self._metrics["total_latency"] / total_queries) * 1000, 2
            ),
            "total_tokens": self._metrics["total_tokens"],
            "avg_tokens_per_query": round(
                self._metrics["total_tokens"] / total_queries, 2
            ),
            "queries_by_model": dict(self._metrics["queries_by_model"]),
            "errors_by_type": dict(self._metrics["errors_by_type"]),
        }

        return metrics

    def get_recent_queries(self, limit: int = 10) -> List[Dict[str, Any]]:
        """
        Get recent query history.

        Args:
            limit: Number of recent queries to return

        Returns:
            list: Recent queries
        """
        return self._query_history[-limit:]

    def reset_metrics(self) -> None:
        """Reset all metrics to zero."""
        self._metrics = {
            "total_queries": 0,
            "successful_queries": 0,
            "failed_queries": 0,
            "cache_hits": 0,
            "cache_misses": 0,
            "total_latency": 0.0,
            "total_tokens": 0,
            "queries_by_model": defaultdict(int),
            "errors_by_type": defaultdict(int),
        }
        self._query_history = []
        self.logger.info("Metrics reset")

    def export_metrics(self, output_file: str) -> None:
        """
        Export metrics to JSON file.

        Args:
            output_file: Path to output file
        """
        metrics = self.get_metrics()
        metrics["timestamp"] = datetime.now().isoformat()
        metrics["recent_queries"] = self.get_recent_queries(20)

        with open(output_file, "w") as f:
            json.dump(metrics, f, indent=2, default=str)

        self.logger.info(f"Metrics exported to {output_file}")

    def log_performance_warning(self, message: str, metric_name: str, value: float, threshold: float) -> None:
        """
        Log performance warning when metrics exceed thresholds.

        Args:
            message: Warning message
            metric_name: Name of the metric
            value: Current value
            threshold: Threshold value
        """
        warning_data = {
            "type": "performance_warning",
            "metric": metric_name,
            "value": value,
            "threshold": threshold,
            "message": message,
            "timestamp": datetime.now().isoformat(),
        }

        self.logger.warning(json.dumps(warning_data))

    def log_error(self, error_type: str, error_message: str, context: Optional[Dict[str, Any]] = None) -> None:
        """
        Log error with context.

        Args:
            error_type: Type of error
            error_message: Error message
            context: Additional context
        """
        error_data = {
            "type": "error",
            "error_type": error_type,
            "message": error_message,
            "timestamp": datetime.now().isoformat(),
        }

        if context:
            error_data["context"] = context

        self.logger.error(json.dumps(error_data))


def test_monitoring():
    """Test the monitoring system."""
    print("=" * 80)
    print("Testing NLQ Monitoring")
    print("=" * 80)

    # Create monitoring instance
    monitor = NLQMonitoring(enable_console=True)

    # Test 1: Log successful query
    print("\nTest 1: Log successful query")
    monitor.log_query(
        query="What is the conversion rate?",
        success=True,
        latency=1.5,
        model="gpt-4",
        tokens_used=150,
        cached=False,
    )

    # Test 2: Log cached query
    print("\nTest 2: Log cached query")
    monitor.log_query(
        query="What is the conversion rate?",
        success=True,
        latency=0.1,
        model="gpt-4",
        tokens_used=0,
        cached=True,
    )

    # Test 3: Log failed query
    print("\nTest 3: Log failed query")
    monitor.log_query(
        query="Invalid query",
        success=False,
        latency=0.5,
        model="gpt-4",
        tokens_used=50,
        cached=False,
        error="ValidationError: Invalid query format",
    )

    # Test 4: Get metrics
    print("\nTest 4: Get metrics")
    metrics = monitor.get_metrics()
    print(json.dumps(metrics, indent=2))

    # Test 5: Get recent queries
    print("\nTest 5: Recent queries")
    recent = monitor.get_recent_queries(limit=3)
    for i, query in enumerate(recent, 1):
        print(f"  Query {i}: {query['query'][:50]} - Success: {query['success']}")

    # Test 6: Performance warning
    print("\nTest 6: Performance warning")
    monitor.log_performance_warning(
        message="High latency detected",
        metric_name="avg_latency_ms",
        value=2500,
        threshold=2000,
    )

    print("\n✅ Monitoring tests completed")
    print("=" * 80)


if __name__ == "__main__":
    test_monitoring()



"""
Semantic Layer for Outhad_Edge Natural Language Query.

Provides business-friendly descriptions of data schema, metrics, and analysis tools.
This layer helps the LLM understand the domain context and generate accurate code.
"""

from typing import Dict, List, Any, Optional
import pandas as pd
from datetime import datetime

try:
    from pydantic import BaseModel
except ImportError:
    # Fallback if pydantic v2 not available
    try:
        from pydantic.v1 import BaseModel
    except ImportError:
        # Basic substitute for very old pydantic
        class BaseModel:
            pass


class ColumnMetadata(BaseModel):
    """Metadata for a single column"""
    name: str
    description: str
    data_type: str
    examples: List[str]
    constraints: Optional[Dict] = None

    class Config:
        arbitrary_types_allowed = True


class TableMetadata(BaseModel):
    """Metadata for a table/dataset"""
    name: str
    description: str
    columns: List[ColumnMetadata]
    relationships: Optional[Dict] = None
    business_terms: List[str] = []

    class Config:
        arbitrary_types_allowed = True


class SemanticLayer:
    """
    Semantic layer that provides business-friendly descriptions
    of data structures for LLM consumption.

    Combines:
    - Schema metadata with pydantic models
    - Business glossary (business terms → technical definitions)
    - API documentation for OuthadEdge operations
    - Runtime data statistics
    """

    def __init__(self, eventstream):
        """
        Initialize semantic layer with an eventstream.

        Args:
            eventstream: Outhad_Edge Eventstream object
        """
        self.eventstream = eventstream
        self.df = eventstream.to_dataframe()
        self.schema = eventstream.schema

        # Initialize metadata
        self.tables = self._initialize_metadata()
        self.business_glossary = self._initialize_glossary()

    def _initialize_metadata(self) -> Dict[str, TableMetadata]:
        """
        Define metadata for OuthadEdge Eventstream
        """
        return {
            'eventstream': TableMetadata(
                name='eventstream',
                description='User event data containing clickstream, behavior, and transaction events',
                columns=[
                    ColumnMetadata(
                        name='user_id',
                        description='Unique identifier for each user',
                        data_type='string',
                        examples=['user_123', 'anon_456'],
                        constraints={'not_null': True}
                    ),
                    ColumnMetadata(
                        name='event',
                        description='Type of event (e.g., page_view, purchase, add_to_cart)',
                        data_type='string',
                        examples=['page_view', 'purchase', 'add_to_cart', 'signup'],
                    ),
                    ColumnMetadata(
                        name='timestamp',
                        description='When the event occurred (datetime)',
                        data_type='datetime',
                        examples=['2024-01-15 10:30:00'],
                    ),
                    ColumnMetadata(
                        name='event_type',
                        description='Category of event (raw, synthetic, processed)',
                        data_type='string',
                        examples=['raw', 'path_start', 'path_end'],
                    ),
                    ColumnMetadata(
                        name='event_index',
                        description='Sequential order of events per user',
                        data_type='integer',
                        examples=['1', '2', '3'],
                    ),
                ],
                business_terms=['users', 'events', 'clickstream', 'behavior', 'journey']
            )
        }

    def _initialize_glossary(self) -> Dict[str, str]:
        """
        Business terms → technical definitions
        """
        return {
            'conversion': 'User completing a purchase or desired action',
            'conversion rate': 'Percentage of users who complete desired action',
            'funnel': 'Sequential steps users take toward a goal',
            'cohort': 'Group of users who share a common characteristic',
            'retention': 'Users who return after initial visit',
            'churn': 'Users who stop using the product',
            'session': 'Single continuous period of user activity',
            'journey': 'Complete path a user takes through the product',
            'touchpoint': 'Interaction point between user and product',
            'mobile': 'Events from mobile devices',
            'desktop': 'Events from desktop computers',
            'anonymous': 'Users without identified credentials',
            'mau': 'Monthly Active Users - unique users in a month',
            'dau': 'Daily Active Users - unique users in a day',
            'ltv': 'Lifetime Value - total revenue from a user',
            'arpu': 'Average Revenue Per User',
        }

    def get_schema_description(self, table: str = 'eventstream') -> str:
        """
        Generate natural language description of schema for LLM
        """
        metadata = self.tables[table]

        description = f"""
Table: {metadata.name}
Description: {metadata.description}

Columns:
"""
        for col in metadata.columns:
            description += f"\n- {col.name} ({col.data_type}): {col.description}"
            if col.examples:
                description += f"\n  Examples: {', '.join(col.examples)}"

        description += f"\n\nBusiness Terms: {', '.join(metadata.business_terms)}"

        return description

    def resolve_business_term(self, term: str) -> Optional[str]:
        """Convert business term to technical definition"""
        return self.business_glossary.get(term.lower())

    def get_eventstream_api_doc(self) -> str:
        """
        Documentation of OuthadEdge Eventstream API for LLM
        """
        return """
# OuthadEdge Eventstream API

## Creating Eventstream
```python
from outhad_edge import Eventstream
stream = Eventstream(df)  # df is pandas DataFrame with columns: user_id, event, timestamp
```

## Common Operations

### Funnel Analysis
```python
funnel = stream.funnel()
funnel.fit(stages=['homepage', 'product', 'cart', 'checkout', 'purchase'])
results = funnel.values  # Get conversion rates
```

### Cohort Analysis
```python
cohorts = stream.cohorts(
    cohort_start_unit='W',  # Weekly cohorts
    cohort_period=(1, 'W')  # 1-week retention periods
)
retention_data = cohorts.cohort_matrix
```

### Transition Graph
```python
graph = stream.transition_graph()
# Returns user flow visualization
```

### Data Access and Filtering
```python
# Get the underlying pandas DataFrame
df = stream.to_dataframe()

# Filter by event name
purchase_events = df[df['event'] == 'purchase']

# Count unique users who converted
converted_users = purchase_events['user_id'].nunique()
total_users = df['user_id'].nunique()

# Calculate conversion rate
conversion_rate = (converted_users / total_users) * 100
```

### User Segmentation
```python
clusters = stream.clusters()
X = clusters.extract_features(method='tfidf')
clusters.fit(method='kmeans', n_clusters=4, X=X)
segments = clusters.cluster_mapping
```

### Statistical Tests
```python
stats = stream.stattests()
result = stats.chi2_test(
    groups=['variant_a', 'variant_b'],
    event='purchase'
)
```

## Available Methods
- stream.describe() - Summary statistics
- stream.user_lifetime_hist() - User lifetime distribution
- stream.event_timestamp_hist() - Event timing patterns
- stream.to_dataframe() - Get underlying pandas DataFrame
"""

    def get_schema_statistics(self) -> Dict[str, Any]:
        """
        Get runtime statistics about the data.

        Returns:
            dict: Schema metadata with actual data statistics
        """
        df = self.df

        # Basic statistics
        total_events = len(df)
        unique_users = df[self.schema.user_id].nunique()

        # Date range
        timestamps = pd.to_datetime(df[self.schema.event_timestamp])
        date_range = f"{timestamps.min()} to {timestamps.max()}"

        # Event distribution
        event_counts = df[self.schema.event_name].value_counts()

        return {
            "user_id_type": str(df[self.schema.user_id].dtype),
            "event_type": str(df[self.schema.event_name].dtype),
            "timestamp_type": str(df[self.schema.event_timestamp].dtype),
            "total_events": total_events,
            "unique_users": unique_users,
            "date_range": date_range,
            "event_counts": event_counts.to_dict(),
            "columns": list(df.columns),
        }

    def get_event_catalog(self) -> List[Dict[str, Any]]:
        """
        Get catalog of all events with descriptions and frequencies.

        Returns:
            list: Event catalog with metadata
        """
        df = self.df
        event_col = self.schema.event_name

        event_stats = (
            df.groupby(event_col)
            .agg(
                {
                    self.schema.user_id: "nunique",
                    event_col: "count",
                }
            )
            .rename(
                columns={
                    self.schema.user_id: "unique_users",
                    event_col: "total_occurrences",
                }
            )
        )

        catalog = []
        for event_name in event_stats.index:
            catalog.append(
                {
                    "event": event_name,
                    "unique_users": int(event_stats.loc[event_name, "unique_users"]),
                    "total_occurrences": int(
                        event_stats.loc[event_name, "total_occurrences"]
                    ),
                    "percentage_of_users": round(
                        event_stats.loc[event_name, "unique_users"] / self.df[self.schema.user_id].nunique() * 100,
                        2,
                    ),
                }
            )

        return sorted(catalog, key=lambda x: x["total_occurrences"], reverse=True)

    def get_tools_description(self) -> str:
        """
        Get description of available Outhad_Edge analysis tools.

        Returns:
            str: Markdown-formatted description of tools
        """
        tools_desc = """
**Available Analysis Tools:**

1. **transition_graph()** - Visualize user flow between events as a network graph
   - Shows transition probabilities between events
   - Interactive visualization
   - Example: `stream.transition_graph()`

2. **funnel(stages=[...])** - Analyze conversion funnel
   - Calculate drop-off rates between stages
   - Supports stage grouping and timeouts
   - Example: `stream.funnel(stages=['view', 'cart', 'purchase'])`

3. **step_matrix(...)** - Step-by-step event distribution analysis
   - Shows event distribution at each path step
   - Can be centered on specific events
   - Example: `stream.step_matrix(max_steps=10)`

4. **cohorts(...)** - Retention cohort analysis
   - Group users by first activity date
   - Track retention over time
   - Example: `stream.cohorts(cohort_start_unit='M', cohort_period=(1, 'M'))`

5. **clusters** - Behavioral segmentation
   - Group users by similar behavior patterns
   - Uses TF-IDF, n-grams, or raw sequences
   - Example: `stream.clusters().fit(method='kmeans', n_clusters=5)`

6. **filter_events(events=[...])** - Filter to specific events
   - Keep only specified events
   - Example: `stream.filter_events(['purchase', 'signup'])`

7. **split_sessions(timeout=...)** - Split paths into sessions
   - Create session boundaries based on time gaps
   - Example: `stream.split_sessions(timeout=(30, 'm'))`

8. **to_dataframe()** - Get raw pandas DataFrame
   - Access underlying data
   - Use for custom pandas operations
"""
        return tools_desc

    def get_sample_data(self, n: int = 5) -> str:
        """
        Get sample rows from the eventstream as formatted string.

        Args:
            n: Number of sample rows

        Returns:
            str: Formatted sample data
        """
        sample = self.df.head(n)[
            [self.schema.user_id, self.schema.event_name, self.schema.event_timestamp]
        ]
        return sample.to_string(index=False)

    def get_full_context(self) -> str:
        """
        Get complete semantic context for LLM.

        Returns:
            str: Full context description combining all metadata
        """
        schema_stats = self.get_schema_statistics()
        event_catalog = self.get_event_catalog()
        tools_desc = self.get_tools_description()
        sample_data = self.get_sample_data()
        api_doc = self.get_eventstream_api_doc()

        # Format event list
        event_list_str = "\n".join(
            [
                f"  - {e['event']}: {e['total_occurrences']:,} occurrences, "
                f"{e['unique_users']:,} users ({e['percentage_of_users']:.1f}%)"
                for e in event_catalog[:15]  # Top 15 events
            ]
        )

        context = f"""
**Eventstream Overview:**
- Total Events: {schema_stats['total_events']:,}
- Unique Users: {schema_stats['unique_users']:,}
- Date Range: {schema_stats['date_range']}
- Columns: {', '.join(schema_stats['columns'])}

**Top Events:**
{event_list_str}

**Sample Data:**
{sample_data}

{tools_desc}

{api_doc}

**Data Access:**
- The eventstream is available as variable `stream`
- Access raw data with: `stream.to_dataframe()`
- Schema columns: user_id='{self.schema.user_id}', event='{self.schema.event_name}', timestamp='{self.schema.event_timestamp}'
"""
        return context

    def to_dict(self) -> Dict[str, Any]:
        """
        Export semantic layer to dictionary.

        Returns:
            dict: Complete semantic layer data
        """
        return {
            "metadata": {
                table: meta.dict() for table, meta in self.tables.items()
            },
            "glossary": self.business_glossary,
            "statistics": self.get_schema_statistics(),
            "events": self.get_event_catalog(),
            "api_documentation": self.get_eventstream_api_doc(),
            "full_context": self.get_full_context(),
        }



from __future__ import annotations

import uuid
from typing import Callable, Optional

from outhad_edge.exceptions.server import ServerNotFoundActionError


class Action:
    def __init__(self, method: str, callback: Callable):
        self.method = method
        self.callback = callback


class JupyterServer:
    pk: str
    actions: dict[str, Action]
    kernel_id: Optional[str]

    def __init__(self, pk: Optional[str] = None, kernel_id: Optional[str] = None):
        self.pk = pk if pk is not None else self.make_id()
        self.actions: dict[str, Action] = {}
        self.kernel_id = kernel_id

    def _find_action(self, method: str) -> None | Action:
        return self.actions.get(method, None)

    def make_id(self) -> str:
        return str(uuid.uuid4())

    def register_action(self, method: str, callback: Callable) -> None:
        self.actions[method] = Action(method, callback)

    def dispatch_method(self, method: str, payload: dict) -> Callable:
        action = self._find_action(method)
        if action is not None:
            return action.callback(payload)
        else:
            raise ServerNotFoundActionError("method not found!", method=method)
